from hashlib import sha256
from json import dumps as json_dumps
from typing import TYPE_CHECKING, Any, Mapping, Type, TypeVar
from urllib.parse import parse_qs, urlencode, urlparse, urlunparse

from fastapi import Response
from pydantic import BaseModel, model_validator
from pydantic._internal._model_construction import ModelMetaclass
from pydantic.fields import Field, FieldInfo
from typing_extensions import dataclass_transform

from mountaineer.client_compiler.build_metadata import BuildMetadata

T = TypeVar("T")


class FieldClassDefinition(BaseModel):
    root_model: Type[BaseModel]
    key: str
    field_definition: FieldInfo

    model_config = {
        "arbitrary_types_allowed": True,
    }


@dataclass_transform(kw_only_default=True, field_specifiers=(Field,))
class ReturnModelMetaclass(ModelMetaclass):
    INTERNAL_RENDER_FIELDS = ["metadata"]

    if not TYPE_CHECKING:  # pragma: no branch
        # Following the lead of the pydantic superclass, we wrap with a non-TYPE_CHECKING
        # block: "otherwise mypy allows arbitrary attribute access""

        def __new__(
            self,
            cls_name: str,
            bases: tuple[type[Any], ...],
            namespace: dict[str, Any],
            *args,
            **kwargs: Any,
        ):
            # Pydantic uses exceptions in the __getattr__ to handle collection of fields
            # in set_model_fields. While we're still initializing the class we have no
            # need for our custom accessor logic - so we temporarily turn it off.
            self.is_constructing = True
            obj = super().__new__(self, cls_name, bases, namespace, *args, **kwargs)
            self.is_constructing = False
            return obj

        def __getattr__(self, key: str) -> Any:
            if self.is_constructing:
                return super().__getattr__(key)

            try:
                return super().__getattr__(key)
            except AttributeError:
                # Determine if this field is defined within the spec
                # If so, return it
                if key in self.model_fields and key not in self.INTERNAL_RENDER_FIELDS:
                    return FieldClassDefinition(
                        root_model=self,
                        key=key,
                        field_definition=self.model_fields[key],
                    )
                raise


class HashableAttribute(BaseModel):
    """
    Even with frozen=True, we can't hash our attributes because they include dictionary field types.
    Instead provide a mixin to calculate the hash of the current state of the attributes.

    """

    def __hash__(self):
        model_json = json_dumps(self.model_dump(), sort_keys=True)
        hash_object = sha256(model_json.encode())
        # __hash__ must return an integer
        return int(hash_object.hexdigest(), 16)


class MetaAttribute(HashableAttribute, BaseModel):
    """
    Represents a meta tag that can be included in the head of an HTML document.

    ```python {{ sticky: True }}
    MetaAttribute(
        name="description",
        content="This is a description of the page.",
        optional_attributes={"lang": "en"},
    )
    ```

    """

    name: str | None = None
    content: str | None = None
    optional_attributes: dict[str, str] = {}


class ThemeColorMeta(MetaAttribute):
    """
    Customizes the default color that is attached to the page.

    ```python {{ sticky: True }}
    ThemeColorMeta(
        color="white",
        media="light",
    )
    ```

    """

    color: str
    media: str | None = None

    @model_validator(mode="after")
    def create_attribute(self):
        self.name = "theme-color"
        self.content = self.color
        if self.media:
            self.optional_attributes = {"media": self.media}
        return self


class ViewportMeta(MetaAttribute):
    """
    Defines the bounds on the current page and how much users are able to zoom.

    ```python {{ sticky: True }}
    ViewportMeta(
        initial_scale=1.0,
        maximum_scale=2.0,
        user_scalable=True,
    )
    ```

    """

    width: str = "device-width"
    initial_scale: float = 1.0
    maximum_scale: float = 1.0
    user_scalable: bool = False

    @model_validator(mode="after")
    def create_attribute(self):
        user_scalable_str = "yes" if self.user_scalable else "no"

        self.name = "viewport"
        self.content = f"width={self.width}, initial-scale={self.initial_scale}, maximum-scale={self.maximum_scale}, user-scalable={user_scalable_str}"
        return self


class LinkAttribute(HashableAttribute, BaseModel):
    """
    Inject a generic <link> tag into the <head> of the current page.

    ```python {{ sticky: True }}
    LinkAttribute(
        rel="stylesheet",
        href="https://example.com/styles.css",
        optional_attributes={"media": "screen"},
    )
    ```

    """

    rel: str
    href: str
    optional_attributes: dict[str, str] = {}

    # A sha will only be added automatically for link imports
    # that reference files built into the _static directory at runtime. Other
    # links will remain as-is without a resolved sha.
    add_static_sha: bool = True

    def set_sha(self, sha: str):
        """
        Updates the URL by adding or modifying the 'sha' query parameter. If a sha
        is already provided as part of href, will override the existing sha.

        """
        # Existing URL components
        parsed_url = urlparse(self.href)
        query_params = parse_qs(parsed_url.query)

        query_params["sha"] = [sha]
        new_query = urlencode(query_params, doseq=True)

        self.href = urlunparse(
            (
                parsed_url.scheme,
                parsed_url.netloc,
                parsed_url.path,
                parsed_url.params,
                new_query,
                parsed_url.fragment,
            )
        )


class ScriptAttribute(HashableAttribute, BaseModel):
    """
    Inject a generic <script> tag into the <head> of the current page.

    ```python {{ sticky: True }}
    ScriptAttribute(
        src="https://example.com/script.js",
        asynchronous=True,
        optional_attributes={"defer": "true"},
        defer=False,
    )
    ```

    """

    src: str
    asynchronous: bool = False
    defer: bool = False
    optional_attributes: dict[str, str] = {}


class Metadata(BaseModel):
    """
    Metadata lets the client specify the different metadata definitions that should
    appear on the current page. These are outside the scope of React management so are
    only handled once on the initial page render. But because they still receive the initial
    render request, you can add any dynamic logic that you want to customize the behavior
    of the page metadata. This includes templating the page title depending on a remote value,
    modifying the scripts injected depending on the user, etc.

    ```python {{ sticky: True }}
    class MyController(Controller):
        async def render(
            self,
            user_name: str,
        ):
            return MyControllerRender(
                metadata=Metadata(
                    title=f"Welcome, {user_name}!",
                )
            )
    ```

    ```python {{ sticky: True }}
    Metadata(
        title="My Page",
        metas=[
            MetaAttribute(...),
            ThemeColorMeta(...),
            ViewportMeta(...),
        ],
        links=[
            LinkAttribute(...),
        ]
    )
    ```

    """

    title: str | None = None

    # Specify dynamic injection of tags into the <head>
    metas: list[MetaAttribute] = []
    links: list[LinkAttribute] = []
    scripts: list[ScriptAttribute] = []

    # Allows the client to specify a different response type
    # that should occur on initial render
    # Useful for redirects, adding cookies, etc.
    explicit_response: Response | None = None

    # If enabled, we won't attempt to use the global metadata for this route
    # Helpful for plugins or otherwise for nested routes that should escape the container
    ignore_global_metadata: bool = False

    model_config = {
        "extra": "forbid",
        "arbitrary_types_allowed": True,
    }

    def merge(self, parent: "Metadata") -> "Metadata":
        def merge_item(a: list[T], b: list[T]):
            # Keeps the original ordering while avoiding duplicates
            for item in b:
                if item not in a:
                    a.append(item)
            return a

        return Metadata(
            title=self.title or parent.title,
            metas=merge_item(self.metas, parent.metas),
            links=merge_item(self.links, parent.links),
            scripts=merge_item(self.scripts, parent.scripts),
            explicit_response=self.explicit_response,
            ignore_global_metadata=self.ignore_global_metadata,
        )

    def build_header(self, build_metadata: BuildMetadata | None) -> list[str]:
        """
        Builds the header for this controller. Returns the list of tags that will be injected into the
        <head> tag of the rendered page.

        """
        tags: list[str] = []

        def format_optional_keys(payload: Mapping[str, str | bool | None]) -> str:
            attributes: list[str] = []
            for key, value in payload.items():
                if value is None:
                    continue
                elif isinstance(value, bool):
                    # Boolean attributes can just be represented by just their key
                    if value:
                        attributes.append(key)
                    else:
                        continue
                else:
                    attributes.append(f'{key}="{value}"')
            return " ".join(attributes)

        if self.title:
            tags.append(f"<title>{self.title}</title>")

        for meta_definition in self.metas:
            meta_attributes = {
                "name": meta_definition.name,
                "content": meta_definition.content,
                **meta_definition.optional_attributes,
            }
            tags.append(f"<meta {format_optional_keys(meta_attributes)} />")

        for script_definition in self.scripts:
            script_attributes: dict[str, str | bool] = {
                "src": script_definition.src,
                "async": script_definition.asynchronous,
                "defer": script_definition.defer,
                **script_definition.optional_attributes,
            }
            tags.append(f"<script {format_optional_keys(script_attributes)}></script>")

        for link_definition in self.links:
            if build_metadata and link_definition.add_static_sha:
                # By convention, static files should be mounted to the application in a /static endpoint. These
                # may be served outside of mountaineer (via a CDN or similar) but we still expect
                # these paths to reference the proper paths
                link_sha: str | None = None
                for (
                    static_path,
                    static_sha,
                ) in build_metadata.static_artifact_shas.items():
                    # Allow for alternative endings to support existing query parameters
                    if link_definition.href.startswith(f"/static/{static_path}"):
                        link_sha = static_sha
                if link_sha:
                    link_definition.set_sha(link_sha)

            link_attributes = {
                "rel": link_definition.rel,
                "href": link_definition.href,
                **link_definition.optional_attributes,
            }

            tags.append(f"<link {format_optional_keys(link_attributes)} />")

        return tags


class RenderBase(BaseModel, metaclass=ReturnModelMetaclass):
    """
    Base class for all renderable data models. Subclass this model when defining
    your own component data schema.
    """

    metadata: Metadata | None = Field(default=None, exclude=True)

    model_config = {
        # Frozen parameters are required so we can hash the render values to check
        # for changes in our SSR renderer
        "frozen": True,
    }


class RenderNull(RenderBase):
    pass



"""
Update the version for the regular Mountaineer rust project.

"""

import sys
from pathlib import Path
from sys import stdout

import toml
from packaging.version import parse


def update_version_rust(new_version: str):
    cargo_path = Path("Cargo.toml")
    if not cargo_path.exists():
        stdout.write("Cargo.toml not found, skipping version update")
        return

    filedata = toml.loads(cargo_path.read_text())

    # If the new version is a pre-release version, we need to reformat it
    # to align with Cargo standards
    # pip format uses "0.1.0.dev1" while Cargo uses "0.1.0-dev1"
    cargo_version = format_cargo_version(new_version)

    if "package" not in filedata:
        raise ValueError("Cargo.toml is missing the [package] section")

    filedata["package"]["version"] = cargo_version

    cargo_path.write_text(toml.dumps(filedata))


def format_cargo_version(new_version: str) -> str:
    parsed_version = parse(new_version)

    cargo_version = (
        f"{parsed_version.major}.{parsed_version.minor}.{parsed_version.micro}"
    )
    if parsed_version.is_prerelease and parsed_version.pre is not None:
        pre_release = ".".join(str(x) for x in parsed_version.pre)
        cargo_version += f"-{pre_release.replace('.', '')}"
    if parsed_version.is_postrelease and parsed_version.post is not None:
        cargo_version += f"-post{parsed_version.post}"
    if parsed_version.is_devrelease and parsed_version.dev is not None:
        cargo_version += f"-dev{parsed_version.dev}"

    return cargo_version


def update_version_python(new_version: str):
    pyproject_path = Path("pyproject.toml")
    if not pyproject_path.exists():
        print("pyproject.toml not found, skipping version update")  # noqa: T201
        return

    filedata = toml.loads(pyproject_path.read_text())

    # Parse the new version to ensure it's valid and potentially reformat
    python_version = format_python_version(new_version)

    updated = False

    # Update poetry version if it exists
    if "tool" in filedata and "poetry" in filedata["tool"]:
        filedata["tool"]["poetry"]["version"] = python_version
        updated = True

    # Update project.version if it exists
    if "project" in filedata and "version" in filedata["project"]:
        filedata["project"]["version"] = python_version
        updated = True

    if not updated:
        print(  # noqa: T201
            "Warning: Neither [tool.poetry] nor [project] sections found in pyproject.toml"
        )
        return

    pyproject_path.write_text(toml.dumps(filedata))


def format_python_version(new_version: str) -> str:
    parsed_version = parse(new_version)
    # Assuming semantic versioning, format it as needed
    python_version = (
        f"{parsed_version.major}.{parsed_version.minor}.{parsed_version.micro}"
    )
    if parsed_version.is_prerelease and parsed_version.pre is not None:
        pre_release = ".".join(str(x) for x in parsed_version.pre)
        python_version += (
            f".dev{parsed_version.pre[-1]}"
            if pre_release.startswith("dev")
            else f"b{parsed_version.pre[-1]}"
        )
    if parsed_version.is_postrelease and parsed_version.post is not None:
        python_version += f".post{parsed_version.post}"
    if parsed_version.is_devrelease and parsed_version.dev is not None:
        python_version += f".dev{parsed_version.dev}"
    return python_version


if __name__ == "__main__":
    if len(sys.argv) != 2:
        stdout.write("Usage: python update_version.py <new_version>")
        sys.exit(1)
    new_version = sys.argv[1].lstrip("v")
    update_version_rust(new_version)
    update_version_python(new_version)
    stdout.write(f"Updated version to: {new_version}")



import pytest

from mountaineer.client_compiler.build_metadata import BuildMetadata
from mountaineer.render import (
    LinkAttribute,
    MetaAttribute,
    Metadata,
    ScriptAttribute,
    ThemeColorMeta,
    ViewportMeta,
)


@pytest.mark.parametrize(
    "metadata, expected_tags",
    [
        # Title
        (
            Metadata(
                title="MyTitle",
            ),
            ["<title>MyTitle</title>"],
        ),
        # Multiple links
        (
            Metadata(
                links=[
                    LinkAttribute(
                        rel="stylesheet",
                        href="/stylesheet1.css",
                    ),
                    LinkAttribute(
                        rel="stylesheet",
                        href="/stylesheet2.css",
                    ),
                ],
            ),
            [
                '<link rel="stylesheet" href="/stylesheet1.css" />',
                '<link rel="stylesheet" href="/stylesheet2.css" />',
            ],
        ),
        # Mixed meta tags
        (
            Metadata(
                metas=[
                    ThemeColorMeta(
                        color="#000000",
                    ),
                    ViewportMeta(
                        width="device-width",
                        initial_scale=1,
                        maximum_scale=2,
                        user_scalable=False,
                    ),
                ]
            ),
            [
                '<meta name="theme-color" content="#000000" />',
                '<meta name="viewport" content="width=device-width, initial-scale=1.0, maximum-scale=2.0, user-scalable=no" />',
            ],
        ),
        # Script tags
        (
            Metadata(
                scripts=[
                    ScriptAttribute(
                        src="/script1.js",
                    ),
                    ScriptAttribute(
                        src="/script2.js",
                        asynchronous=True,
                    ),
                    ScriptAttribute(
                        src="/script3.js",
                        defer=True,
                    ),
                    ScriptAttribute(
                        src="/script4.js",
                        optional_attributes={"test-attr": "test-value"},
                    ),
                ],
            ),
            [
                '<script src="/script1.js"></script>',
                '<script src="/script2.js" async></script>',
                '<script src="/script3.js" defer></script>',
                '<script src="/script4.js" test-attr="test-value"></script>',
            ],
        ),
    ],
)
def test_build_header(metadata: Metadata, expected_tags: list[str]):
    assert metadata.build_header(build_metadata=None) == expected_tags


@pytest.mark.parametrize(
    "metadata, build_metadata, expected_tags",
    [
        # Test static file with matching SHA
        (
            Metadata(
                links=[
                    LinkAttribute(
                        rel="stylesheet",
                        href="/static/css/style.css",
                        add_static_sha=True,
                    )
                ]
            ),
            BuildMetadata(static_artifact_shas={"css/style.css": "abc123"}),
            ['<link rel="stylesheet" href="/static/css/style.css?sha=abc123" />'],
        ),
        # Test multiple static files with matching SHAs
        (
            Metadata(
                links=[
                    LinkAttribute(
                        rel="stylesheet",
                        href="/static/css/style.css",
                        add_static_sha=True,
                    ),
                    LinkAttribute(
                        rel="stylesheet",
                        href="/static/css/other.css",
                        add_static_sha=True,
                    ),
                ]
            ),
            BuildMetadata(
                static_artifact_shas={
                    "css/style.css": "abc123",
                    "css/other.css": "def456",
                }
            ),
            [
                '<link rel="stylesheet" href="/static/css/style.css?sha=abc123" />',
                '<link rel="stylesheet" href="/static/css/other.css?sha=def456" />',
            ],
        ),
        # Test static file without matching SHA (should not modify URL)
        (
            Metadata(
                links=[
                    LinkAttribute(
                        rel="stylesheet",
                        href="/static/css/nonexistent.css",
                        add_static_sha=True,
                    )
                ]
            ),
            BuildMetadata(static_artifact_shas={"css/style.css": "abc123"}),
            ['<link rel="stylesheet" href="/static/css/nonexistent.css" />'],
        ),
        # Test mix of static and non-static files
        (
            Metadata(
                links=[
                    LinkAttribute(
                        rel="stylesheet",
                        href="/static/css/style.css",
                        add_static_sha=True,
                    ),
                    LinkAttribute(
                        rel="stylesheet", href="/css/external.css", add_static_sha=True
                    ),
                ]
            ),
            BuildMetadata(static_artifact_shas={"css/style.css": "abc123"}),
            [
                '<link rel="stylesheet" href="/static/css/style.css?sha=abc123" />',
                '<link rel="stylesheet" href="/css/external.css" />',
            ],
        ),
        # Test with add_static_sha=False
        (
            Metadata(
                links=[
                    LinkAttribute(
                        rel="stylesheet",
                        href="/static/css/style.css",
                        add_static_sha=False,
                    )
                ]
            ),
            BuildMetadata(static_artifact_shas={"css/style.css": "abc123"}),
            ['<link rel="stylesheet" href="/static/css/style.css" />'],
        ),
        # Test with no build_metadata
        (
            Metadata(
                links=[
                    LinkAttribute(
                        rel="stylesheet",
                        href="/static/css/style.css",
                        add_static_sha=True,
                    )
                ]
            ),
            None,
            ['<link rel="stylesheet" href="/static/css/style.css" />'],
        ),
        # Test with existing query parameters
        (
            Metadata(
                links=[
                    LinkAttribute(
                        rel="stylesheet",
                        href="/static/css/style.css?v=1.0",
                        add_static_sha=True,
                    )
                ]
            ),
            BuildMetadata(static_artifact_shas={"css/style.css": "abc123"}),
            ['<link rel="stylesheet" href="/static/css/style.css?v=1.0&sha=abc123" />'],
        ),
        # Test with mixed attributes and SHA
        (
            Metadata(
                links=[
                    LinkAttribute(
                        rel="stylesheet",
                        href="/static/css/style.css",
                        add_static_sha=True,
                        optional_attributes={
                            "media": "screen",
                            "crossorigin": "anonymous",
                        },
                    )
                ]
            ),
            BuildMetadata(static_artifact_shas={"css/style.css": "abc123"}),
            [
                '<link rel="stylesheet" href="/static/css/style.css?sha=abc123" media="screen" crossorigin="anonymous" />'
            ],
        ),
    ],
)
def test_build_header_with_sha(
    metadata: Metadata, build_metadata: BuildMetadata | None, expected_tags: list[str]
):
    """
    Test the SHA addition logic for static files in the build_header method.

    """
    assert metadata.build_header(build_metadata=build_metadata) == expected_tags


COMPLEX_METADATA = Metadata(
    title="MyTitle",
    links=[
        LinkAttribute(
            rel="stylesheet",
            href="/stylesheet1.css",
        )
    ],
    metas=[
        MetaAttribute(
            name="theme-color",
            content="#000000",
        )
    ],
    scripts=[
        ScriptAttribute(
            src="/script1.js",
        )
    ],
)


@pytest.mark.parametrize(
    "metadatas, expected_metadata",
    [
        # A complex metadata definition should always echo its own definition
        ([COMPLEX_METADATA], COMPLEX_METADATA),
        # We shouldn't end up with duplicates
        (
            [
                COMPLEX_METADATA,
                COMPLEX_METADATA,
            ],
            COMPLEX_METADATA,
        ),
        # Test a simple merge of two different values on the same property
        (
            [
                Metadata(
                    links=[
                        LinkAttribute(
                            rel="stylesheet",
                            href="/stylesheet1.css",
                        )
                    ]
                ),
                Metadata(
                    links=[
                        LinkAttribute(
                            rel="stylesheet",
                            href="/stylesheet2.css",
                        )
                    ]
                ),
            ],
            Metadata(
                links=[
                    LinkAttribute(
                        rel="stylesheet",
                        href="/stylesheet1.css",
                    ),
                    LinkAttribute(
                        rel="stylesheet",
                        href="/stylesheet2.css",
                    ),
                ]
            ),
        ),
        # The first specified header should win.
        (
            [
                Metadata(title="Primary"),
                Metadata(title="Secondary"),
            ],
            Metadata(title="Primary"),
        ),
    ],
)
def test_merge_metadatas(metadatas: list[Metadata], expected_metadata: Metadata):
    metadata = metadatas[0]
    for other_metadata in metadatas[1:]:
        metadata = metadata.merge(other_metadata)

    assert metadata == expected_metadata


@pytest.mark.parametrize(
    "initial_url,new_sha,expected_url",
    [
        # No existing query parameters
        ("https://example.com/path", "abc123", "https://example.com/path?sha=abc123"),
        # Has existing sha parameter
        (
            "https://example.com/path?sha=old123",
            "new456",
            "https://example.com/path?sha=new456",
        ),
        # Has other query parameters but no sha
        (
            "https://example.com/path?param1=value1&param2=value2",
            "def789",
            "https://example.com/path?param1=value1&param2=value2&sha=def789",
        ),
        # Has multiple query parameters including sha
        (
            "https://example.com/path?param1=value1&sha=old123&param2=value2",
            "ghi012",
            "https://example.com/path?param1=value1&sha=ghi012&param2=value2",
        ),
        # URL with special characters
        (
            "https://example.com/path?param=special+value&sha=old",
            "jkl345",
            "https://example.com/path?param=special+value&sha=jkl345",
        ),
        # URL with fragment
        (
            "https://example.com/path?param=value#fragment",
            "mno678",
            "https://example.com/path?param=value&sha=mno678#fragment",
        ),
        # URL with empty query string
        ("https://example.com/path?", "pqr901", "https://example.com/path?sha=pqr901"),
        # URL with port number
        (
            "https://example.com:8080/path?param=value",
            "stu234",
            "https://example.com:8080/path?param=value&sha=stu234",
        ),
    ],
)
def test_link_attribute_set_sha(initial_url, new_sha, expected_url):
    link = LinkAttribute(rel="test", href=initial_url)
    link.set_sha(new_sha)
    assert link.href == expected_url



from contextlib import asynccontextmanager
from pathlib import Path
from time import monotonic_ns
from typing import Any
from unittest.mock import patch

import pytest
from fastapi import Depends, Request
from fastapi.testclient import TestClient
from pydantic.main import BaseModel
from starlette.datastructures import Headers

from mountaineer.__tests__.common import calculate_primes
from mountaineer.actions.fields import FunctionActionType, get_function_metadata
from mountaineer.actions.sideeffect_dec import sideeffect
from mountaineer.annotation_helpers import MountaineerUnsetValue
from mountaineer.app import AppController
from mountaineer.controller import ControllerBase
from mountaineer.controller_layout import LayoutControllerBase
from mountaineer.logging import LOGGER
from mountaineer.render import RenderBase


class ExampleRenderModel(RenderBase):
    value_a: str
    value_b: str


def test_markup_sideeffect():
    """
    Check that the @sideeffect decorator extracts the expected
    data from our model definition.
    """

    class ExamplePassthroughModel(BaseModel):
        first_name: str

    class TestController(ControllerBase):
        view_path = "/test.tsx"

        # We declare as complicated a payload as @sideeffect supports so we can
        # see the full amount of metadata properties that are set
        @sideeffect(
            reload=tuple([ExampleRenderModel.value_a]),
        )
        def sideeffect_and_return_data(self) -> ExamplePassthroughModel:
            return ExamplePassthroughModel(
                first_name="John",
            )

    metadata = get_function_metadata(TestController.sideeffect_and_return_data)
    assert metadata.action_type == FunctionActionType.SIDEEFFECT
    assert metadata.get_passthrough_model() == ExamplePassthroughModel
    assert metadata.function_name == "sideeffect_and_return_data"
    assert metadata.reload_states == tuple([ExampleRenderModel.value_a])
    assert isinstance(metadata.render_model, MountaineerUnsetValue)


class ControllerCommon(ControllerBase):
    url: str = "/test/{query_id}/"
    view_path = "/test.tsx"

    def __init__(self):
        super().__init__()
        self.counter = 0
        self.render_counts = 0

    @sideeffect
    def call_sideeffect(self, payload: dict) -> None:
        self.counter += 1

    @sideeffect
    async def call_sideeffect_async(self, payload: dict) -> None:
        self.counter += 1


async def call_sideeffect_common(controller: ControllerCommon):
    app = AppController(view_root=Path())
    app.register(controller)

    @asynccontextmanager
    async def mock_get_render_parameters(*args, **kwargs):
        yield {
            "query_id": 1,
        }

    # After our wrapper is called, our function is now async
    # Avoid the dependency resolution logic since that's tested separately
    with patch(
        "mountaineer.actions.sideeffect_dec.get_render_parameters"
    ) as patched_get_render_params:
        patched_get_render_params.side_effect = mock_get_render_parameters

        # Even if the "request" is not required by our sideeffects, it's required
        # by the function injected by the sideeffect decorator.
        return_value_sync = await controller.call_sideeffect(  # type: ignore
            {},
            request=Request({"type": "http"}),  # type: ignore
        )

        return_value_async = await controller.call_sideeffect_async(  # type: ignore
            {},
            request=Request({"type": "http"}),  # type: ignore
        )

        # The response payload should be the same both both sync and async endpoints
        expected_response = {
            "sideeffect": ExampleRenderModel(
                value_a="Hello",
                value_b="World",
            ),
            "passthrough": None,
        }

        assert return_value_sync == expected_response
        assert return_value_async == expected_response

        assert controller.counter == 2
        assert controller.render_counts == 2


@pytest.mark.asyncio
async def test_can_call_sideeffect():
    """
    Ensure that we can call the sideeffect, which will in turn
    call the render function to get fresh data.
    """

    class ExampleController(ControllerCommon):
        def render(
            self,
            query_id: int,
        ) -> ExampleRenderModel:
            self.render_counts += 1
            return ExampleRenderModel(
                value_a="Hello",
                value_b="World",
            )

    await call_sideeffect_common(ExampleController())


@pytest.mark.asyncio
async def test_can_call_sideeffect_async_render():
    """
    Render functions can also work asynchronously.
    """

    class ExampleController(ControllerCommon):
        async def render(
            self,
            query_id: int,
        ) -> ExampleRenderModel:
            self.render_counts += 1
            return ExampleRenderModel(
                value_a="Hello",
                value_b="World",
            )

    await call_sideeffect_common(ExampleController())


@pytest.mark.asyncio
async def test_can_call_sideeffect_original():
    """
    Ensure that we can access the raw underlying function that was
    wrapped by the decorator.

    """

    class ExampleController(ControllerCommon):
        def render(
            self,
            query_id: int,
        ) -> ExampleRenderModel:
            self.render_counts += 1
            return ExampleRenderModel(
                value_a="Hello",
                value_b="World",
            )

    controller = ExampleController()
    await ExampleController.call_sideeffect.original(controller, dict())
    await ExampleController.call_sideeffect_async.original(controller, dict())


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "referer, expected_resolved",
    [
        # Fully specified query parameters
        (
            "http://example.com/test/5/?url_query_param=test-query-value",
            {
                "cookie_dependency": "cookie-value",
                "path_param": 5,
                "url_query_param": "test-query-value",
            },
        ),
        # Unspecified query parameters should be None
        (
            "http://example.com/test/5/",
            {
                "cookie_dependency": "cookie-value",
                "path_param": 5,
                "url_query_param": None,
            },
        ),
        # Partially specified query param url
        (
            "http://example.com/test/5/?url_query_param=",
            {
                "cookie_dependency": "cookie-value",
                "path_param": 5,
                "url_query_param": "",
            },
        ),
    ],
)
async def test_get_render_parameters(
    referer: str,
    expected_resolved: dict[str, Any],
):
    """
    Given a controller, reproduce the logic of FastAPI to sniff the render()
    function for dependencies and resolve them.

    """
    from mountaineer.actions.sideeffect_dec import get_render_parameters

    found_cookie = None

    def grab_cookie_dependency(request: Request):
        nonlocal found_cookie
        found_cookie = request.cookies.get("test-cookie")
        return found_cookie

    class TestController(ControllerBase):
        url: str = "/test/{path_param}/"
        view_path = "/test.tsx"

        def render(
            self,
            path_param: int,
            url_query_param: str | None = None,
            cookie_dependency: str = Depends(grab_cookie_dependency),
        ) -> ExampleRenderModel:
            return ExampleRenderModel(
                value_a="Hello",
                value_b="World",
            )

    # We need to load this test controller to an actual application runtime
    # or else we don't have the render() metadata added
    app = AppController(view_root=Path())
    controller = TestController()
    app.register(controller)

    fake_request = Request(
        {
            "type": "http",
            "headers": Headers(
                {
                    "cookie": "test-cookie=cookie-value",
                    # Its important the referer aligns with the controller url, since that is expected
                    # to be the original view page that is calling this sub-function
                    "referer": referer,
                }
            ).raw,
            "http_version": "1.1",
            "scheme": "",
            "client": "",
            "server": "",
            # The URL and method should both be different, to test whether we are able
            # to map the request to the correct endpoint
            "method": "POST",
            "url": "http://localhost/related_action_endpoint",
        }
    )

    async with get_render_parameters(controller, fake_request) as resolved_dependencies:
        assert resolved_dependencies == expected_resolved


@pytest.mark.parametrize(
    "use_experimental,min_time,max_time",
    [
        (False, 1, None),
        (True, None, 0.10),
    ],
)
def test_limit_codepath_experimental(
    use_experimental: bool,
    min_time: float | None,
    max_time: float | None,
):
    class ExampleController(ControllerBase):
        url: str = "/test/{query_id}/"
        view_path = "/test.tsx"

        def render(
            self,
            query_id: int,
        ) -> ExampleRenderModel:
            a = calculate_primes(10000)
            b = calculate_primes(1000000)
            return ExampleRenderModel(
                value_a=f"Hello {a}",
                value_b=f"World {b}",
            )

        @sideeffect(
            reload=(ExampleRenderModel.value_a,),
            experimental_render_reload=use_experimental,
        )
        def call_sideeffect(self, payload: dict) -> None:
            pass

    # We need to load this test controller to an actual application runtime
    # or else we don't have the render() metadata added
    app = AppController(view_root=Path())
    controller = ExampleController()
    app.register(controller)

    controller_definition = app._definition_for_controller(controller)
    sideeffect_url = controller_definition.get_url_for_metadata(
        get_function_metadata(ExampleController.call_sideeffect)
    )

    client = TestClient(app.app)
    start = monotonic_ns()
    response = client.post(
        sideeffect_url,
        json={},
        headers={
            # From the original view page that is calling this sub-function
            "referer": "http://example.com/test/5/",
        },
    )
    elapsed = (monotonic_ns() - start) / 1e9
    assert response.status_code == 200
    assert response.json() == {
        "sideeffect": {
            "value_a": "Hello 1229",
        }
    }

    LOGGER.info(f"Use Experimental: {use_experimental}\nElapsed: {elapsed}")

    if min_time is not None:
        assert elapsed >= min_time
    if max_time is not None:
        assert elapsed <= max_time


@pytest.mark.asyncio
async def test_layout_controller_request_support():
    """
    Test that LayoutControllerBase controllers can now access request information in their render methods
    when called through a sideeffect.

    This was previously not supported, but recent changes now allow request information
    to be passed to LayoutControllerBase render methods.
    """

    class ExampleLayoutRender(RenderBase):
        request_info: str
        cookie_value: str

    class TestLayoutController(LayoutControllerBase):
        view_path = "/test-layout.tsx"

        def render(
            self,
            request: Request,
        ) -> ExampleLayoutRender:
            return ExampleLayoutRender(
                request_info=request.url.path,
                cookie_value=request.cookies.get("test-cookie", "no-cookie"),
            )

        @sideeffect
        def call_sideeffect(self, payload: dict) -> None:
            pass

    # Setup the app controller with our layout controller
    app = AppController(view_root=Path())
    controller = TestLayoutController()
    app.register(controller)

    controller_definition = app._definition_for_controller(controller)
    sideeffect_url = controller_definition.get_url_for_metadata(
        get_function_metadata(TestLayoutController.call_sideeffect)
    )

    # Create a test client and make the request
    client = TestClient(app.app)
    response = client.post(
        sideeffect_url,
        json={},
        headers={
            "Cookie": "test-cookie=layout-cookie-value",
            "referer": "http://example.com/test-page/",
        },
    )

    # Verify the response contains the request information
    assert response.status_code == 200
    response_data = response.json()

    assert "sideeffect" in response_data
    assert response_data["sideeffect"]["request_info"] == "/test-page/"
    assert response_data["sideeffect"]["cookie_value"] == "layout-cookie-value"


@pytest.mark.asyncio
async def test_controller_and_layout_request_handling():
    """
    Test and compare both ControllerBase and LayoutControllerBase to ensure they both
    properly handle request information when called through a sideeffect.

    This test validates the fix that allows layout controllers to access request information,
    which was previously unsupported.
    """

    class ExampleRenderWithRequest(RenderBase):
        path: str
        query_param: str | None
        cookie_value: str
        controller_type: str

    # Standard controller implementation
    class StandardController(ControllerBase):
        url: str = "/standard/{path_param}/"
        view_path = "/standard.tsx"

        def render(
            self,
            path_param: str,
            request: Request,
            query_param: str | None = None,
        ) -> ExampleRenderWithRequest:
            return ExampleRenderWithRequest(
                path=request.url.path,
                query_param=query_param,
                cookie_value=request.cookies.get("test-cookie", "no-cookie"),
                controller_type="standard",
            )

        @sideeffect
        def standard_sideeffect(self, request: Request, payload: dict) -> None:
            pass

    # Layout controller implementation
    class LayoutController(LayoutControllerBase):
        view_path = "/layout.tsx"

        def render(
            self,
            request: Request,
        ) -> ExampleRenderWithRequest:
            return ExampleRenderWithRequest(
                path=request.url.path,
                query_param=request.query_params.get("query_param"),
                cookie_value=request.cookies.get("test-cookie", "no-cookie"),
                controller_type="layout",
            )

        @sideeffect
        def layout_sideeffect(self, request: Request, payload: dict) -> None:
            pass

    # Setup app with both controllers
    app = AppController(view_root=Path())
    standard_controller = StandardController()
    layout_controller = LayoutController()
    app.register(standard_controller)
    app.register(layout_controller)

    # Get the sideeffect URLs for both controllers
    standard_definition = app._definition_for_controller(standard_controller)
    layout_definition = app._definition_for_controller(layout_controller)

    standard_sideeffect_url = standard_definition.get_url_for_metadata(
        get_function_metadata(StandardController.standard_sideeffect)
    )
    layout_sideeffect_url = layout_definition.get_url_for_metadata(
        get_function_metadata(LayoutController.layout_sideeffect)
    )

    # Create a test client and make requests to both controllers
    client = TestClient(app.app)

    # Test standard controller
    client.post(
        standard_sideeffect_url,
        json={},
        headers={
            "Cookie": "test-cookie=standard-cookie-value",
            # Don't include query params in the referer - we'll use the mock_get_render_parameters
            "referer": "http://example.com/standard/param/",
        },
    )

    # Test layout controller
    client.post(
        layout_sideeffect_url,
        json={},
        headers={
            "Cookie": "test-cookie=layout-cookie-value",
            "referer": "http://example.com/some-page/",
        },
    )

    # Use a mock to provide the query parameters correctly
    with patch(
        "mountaineer.actions.sideeffect_dec.get_render_parameters"
    ) as mock_get_params:
        # For standard controller
        @asynccontextmanager
        async def mock_standard_params(*args, **kwargs):
            yield {
                "path_param": "param",
                "request": Request(
                    {
                        "type": "http",
                        "path": "/standard/param/",
                        "query_string": b"query_param=test-value",
                        "headers": Headers(
                            {"cookie": "test-cookie=standard-cookie-value"}
                        ).raw,
                        "method": "GET",
                        "scheme": "http",
                        "client": ("testclient", 50000),
                        "server": ("testserver", 80),
                    }
                ),
                "query_param": "test-value",
            }

        # For layout controller
        @asynccontextmanager
        async def mock_layout_params(*args, **kwargs):
            yield {
                "request": Request(
                    {
                        "type": "http",
                        "path": "/some-page/",
                        "query_string": b"query_param=layout-value",
                        "headers": Headers(
                            {"cookie": "test-cookie=layout-cookie-value"}
                        ).raw,
                        "method": "GET",
                        "scheme": "http",
                        "client": ("testclient", 50000),
                        "server": ("testserver", 80),
                    }
                ),
            }

        # First call uses standard params, second uses layout params
        mock_get_params.side_effect = [mock_standard_params(), mock_layout_params()]

        # Call sideeffects directly to use our mocks
        standard_sideeffect_result = await standard_controller.standard_sideeffect(
            payload={},
            request=Request({"type": "http"}),
        )

        layout_sideeffect_result = await layout_controller.layout_sideeffect(
            payload={},
            request=Request({"type": "http"}),
        )

    # Verify standard controller result
    assert "sideeffect" in standard_sideeffect_result
    sideeffect_model = standard_sideeffect_result["sideeffect"]
    assert isinstance(sideeffect_model, ExampleRenderWithRequest)
    assert sideeffect_model.path == "/standard/param/"
    assert sideeffect_model.query_param == "test-value"
    assert sideeffect_model.cookie_value == "standard-cookie-value"
    assert sideeffect_model.controller_type == "standard"

    # Verify layout controller result
    assert "sideeffect" in layout_sideeffect_result
    layout_model = layout_sideeffect_result["sideeffect"]
    assert isinstance(layout_model, ExampleRenderWithRequest)
    assert layout_model.path == "/some-page/"
    assert layout_model.query_param == "layout-value"
    assert layout_model.cookie_value == "layout-cookie-value"
    assert layout_model.controller_type == "layout"


@pytest.mark.asyncio
async def test_layout_controller_session_handling():
    """
    Test that layout controllers can properly access session information from requests.

    This test specifically focuses on session handling, which is a key aspect of the recent fix
    in the @sideeffect decorator for layout controllers.
    """

    class SessionRenderModel(RenderBase):
        has_session: bool
        session_value: str | None

    class SessionLayoutController(LayoutControllerBase):
        view_path = "/session-layout.tsx"

        def render(
            self,
            request: Request,
        ) -> SessionRenderModel:
            # Check if session exists and get value from it
            session = request.scope.get("session")
            return SessionRenderModel(
                has_session=session is not None,
                session_value=session.get("test_key") if session else None,
            )

        @sideeffect
        def session_sideeffect(self, request: Request, payload: dict) -> None:
            pass

    # Setup app with session middleware
    app = AppController(view_root=Path())
    controller = SessionLayoutController()
    app.register(controller)

    # Get sideeffect URL
    controller_definition = app._definition_for_controller(controller)
    sideeffect_url = controller_definition.get_url_for_metadata(
        get_function_metadata(SessionLayoutController.session_sideeffect)
    )

    # Create a test client
    client = TestClient(app.app)

    # Custom middleware to inject a session into the request
    from starlette.middleware import Middleware
    from starlette.middleware.base import BaseHTTPMiddleware

    class SessionMiddleware(BaseHTTPMiddleware):
        async def dispatch(self, request, call_next):
            request.scope["session"] = {"test_key": "session_test_value"}
            return await call_next(request)

    # Apply middleware manually
    app.app.user_middleware = [Middleware(SessionMiddleware)]
    app.app.build_middleware_stack()

    # Make request with session
    response = client.post(
        sideeffect_url,
        json={},
        headers={
            "referer": "http://example.com/some-page/",
        },
    )

    # Verify session was properly accessed
    assert response.status_code == 200
    response_data = response.json()

    assert response_data["sideeffect"]["has_session"]
    assert response_data["sideeffect"]["session_value"] == "session_test_value"



{% if create_stub_files %}
from mountaineer import Metadata, RenderBase, ControllerBase, APIException, sideeffect
from iceaxe import DBConnection
from iceaxe.mountaineer import DatabaseDependencies

from fastapi import Depends
from pydantic import BaseModel

from {{project_name}} import models

class NotFoundException(APIException):
    status_code = 404
    detail = "Detail item not found"


class UpdateTextRequest(BaseModel):
    description: str


class DetailRender(RenderBase):
    id: int
    description: str


class DetailController(ControllerBase):
    url = "/detail/{detail_id}/"
    view_path = "/app/detail/page.tsx"

    def __init__(self):
        super().__init__()

    async def render(
        self,
        detail_id: int,
        session: DBConnection = Depends(DatabaseDependencies.get_db_connection)
    ) -> DetailRender:
        detail_item = await session.get(models.DetailItem, detail_id)
        if not detail_item:
            raise NotFoundException()

        return DetailRender(
            id=detail_item.id,
            description=detail_item.description,
            metadata=Metadata(title=f"Detail: {detail_id}"),
        )

    @sideeffect
    async def update_text(
        self,
        detail_id: int,
        payload: UpdateTextRequest,
        session: DBConnection = Depends(DatabaseDependencies.get_db_connection)
    ) -> None:
        detail_item = await session.get(models.DetailItem, detail_id)
        if not detail_item:
            raise NotFoundException()

        detail_item.description = payload.description
        await session.update([detail_item])
{% endif %}



from abc import ABC
from importlib.metadata import PackageNotFoundError
from inspect import getmembers, isfunction, ismethod
from pathlib import Path
from re import compile as re_compile
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Coroutine,
    Generic,
    Iterable,
    Optional,
    ParamSpec,
)

from inflection import underscore

from mountaineer.actions import (
    FunctionActionType,
    FunctionMetadata,
    get_function_metadata,
)
from mountaineer.client_compiler.source_maps import SourceMapParser
from mountaineer.config import get_config
from mountaineer.constants import DEFAULT_STATIC_DIR
from mountaineer.logging import LOGGER
from mountaineer.paths import ManagedViewPath, resolve_package_path
from mountaineer.render import (
    RenderBase,
)

if TYPE_CHECKING:
    from mountaineer.app import ControllerDefinition

RenderInput = ParamSpec("RenderInput")


class ControllerBase(ABC, Generic[RenderInput]):
    """
    One Controller should be created for every frontend page in your webapp. The controller
    is where you place all logic that's necessary for this one page - the data that's pushed
    from the frontend and the data that's received from the page.

    All data from `render` is passed to the frontend and usable with automatically
    generated typehints.

    ```python {{sticky: True}}
    from mountaineer import ControllerBase, RenderBase

    class MyControllerRender(RenderBase):
        value: int

    class MyController(ControllerBase):
        url = "/my-page"
        view_path = "/app/my-page.tsx"

        async def render(self) -> MyControllerRender:
            return MyControllerRender(value=10)
    ```

    ```typescript {{sticky: True}}
    import { useServer } from "./_server";

    const MyPage = () => {
        const serverState = useServer();
        return <div>{serverState.value}</div>;
    }

    export default MyPage;
    ```

    """

    url: str
    """
    The URL that this controller will be mounted at. This can contain dynamic
    path parameters, e.g. `/user/{user_id}`. Each parameter will be passed to your
    render function as a keyword argument, so this function would
    have a signature like `async def render(self, user_id: str) -> RenderBase`.

    """

    view_path: str | ManagedViewPath
    """
    Typically, view paths should be a relative path to the local project root.
    Paths are only used if you need to specify an absolute path to another
    file on disk.

    """

    _bundled_scripts: list[str]
    """
    Client static scripts that are identified at runtime. Intended
    for internal use.

    """

    _scripts_prefix: str = DEFAULT_STATIC_DIR
    """
    Prefix for the static resolution endpoint of the bundled scripts.
    """

    _definition: Optional["ControllerDefinition"] = None
    """
    Upon registration, the AppController will mount a wrapper
    with state metadata. This is a back-reference to allow clients
    to access the definition directly from the controller. Intended
    for internal use.

    """

    slow_ssr_threshold: float
    """
    If a server-side rendering operation takes longer than this threshold,
    we will log the time and path parameters as a warning to help debugging.

    """

    hard_ssr_timeout: float | None
    """
    If a server-side rendering operation takes longer than this threshold,
    we will automatically kill the V8 runtime and return an error to the client.
    This helps avoid blocking other server render handlers if the React render
    logic hangs.

    """

    source_map: SourceMapParser | None
    """
    During development, we will load server-side source maps alongside the raw
    javascript code. This parser controls converting stack traces from the
    minified code to the original source code.

    """

    def __init__(
        self, slow_ssr_threshold: float = 0.1, hard_ssr_timeout: float | None = 10.0
    ):
        """
        Clients can override this `__init__` function so long as they call `super().__init__()` at
        the start of their init to setup the internal handlers.

        :param slow_ssr_threshold: Each python process has a single V8 runtime associated with
        it, so SSR rendering can become a bottleneck if it requires processing. We log a warning
        if we detect that an SSR render took longer than this threshold.
        :param hard_ssr_timeout: If the SSR render takes longer than this threshold, we will
        automatically kill the V8 runtime and return an error to the client. This is useful for
        avoiding blocking the reset of the server process if the React render logic hangs.

        """
        super().__init__()

        # Injected by the build framework
        self._bundled_scripts: list[str] = []
        self.slow_ssr_threshold = slow_ssr_threshold
        self.hard_ssr_timeout = hard_ssr_timeout
        self.source_map: SourceMapParser | None = None
        self.initialized = True

        # Set by the path resolution layer
        self._view_base_path: Path | None = None
        self._ssr_path: Path | None = None

        self.resolve_paths()

    def render(
        self, *args: RenderInput.args, **kwargs: RenderInput.kwargs
    ) -> RenderBase | None | Coroutine[Any, Any, RenderBase | None]:
        """
        Render provides the raw data payload that will be sent to the frontend on initial
        render and during any sideeffect update. In most cases, you should return a RenderBase
        instance. If you have no data to display you can also return None.

        This function must be explicitly typehinted with your response type, which allows the
        AppController to generate the correct TypeScript types for the frontend:

        ```python
        class MyServerData(RenderBase):
            pass

        class MyController:
            def render(self) -> MyServerData:
                pass
        ```

        If you don't intend to sync any data from server->client you can typehint this function
        with an explicit None return annotation:

        ```python
        class MyController:
            def render(self) -> None:
                pass
        ```

        Render functions accept any number of arguments and keyword arguments, following the FastAPI
        route parameter style. This includes query parameters, path parameters, and request bodies.

        ```python
        class MyController:
            url = "/my-url/{path_param}"

            def render(
                self,
                query_param: str,
                path_param: int,
                dependency: MyDependency = Depends(MyDependency),
            ) -> MyServerData:
                ...
        ```

        :return: A RenderBase instance or None
        """
        pass

    def _get_client_functions(self) -> Iterable[tuple[str, Callable, FunctionMetadata]]:
        """
        Returns all of the client-callable functions for this controller. Right now we force
        client accessible functions to either be wrapped by @sideeffect or @passthrough.

        """
        # Iterate over all the functions in this class and see which ones have a _metadata attribute
        # We specifically traverse through the MRO, except the last one (object class)
        for name, func in getmembers(self, predicate=ismethod):
            yield from function_is_action(name, func)

    def resolve_paths(self, view_base: Path | None = None, force: bool = True) -> bool:
        """
        Typically used internally by the Mountaineer build pipeline. Calling this function
        sets the active `view_base` of the frontend project, which allows us to resolve the
        built javascripts that are required for this controller.

        :return: Whether we have found all necessary files and fully updated the controller state.

        """
        if not force and self._view_base_path is not None:
            return False

        # Try to resolve the view base path from the global config
        if view_base is None:
            try:
                config = get_config()
                if config.PACKAGE:
                    view_base = Path(resolve_package_path(config.PACKAGE)) / "views"
            except (ValueError, PackageNotFoundError):
                # Config isn't registered yet
                pass

        if view_base is None:
            # Unable to resolve, no-op
            return False

        self._view_base_path = view_base

        # We'll update this bool if we can't find any dependencies
        found_dependencies = True

        # The SSR path is going to be static
        ssr_path = view_base / "_ssr" / f"{self.script_name}.js"
        if ssr_path.exists():
            self._ssr_path = ssr_path
            ssr_map_path = ssr_path.with_suffix(".js.map")
            self.source_map = (
                SourceMapParser(ssr_map_path) if ssr_map_path.exists() else None
            )
        else:
            found_dependencies = False

        # Find the md5-converted cache path
        md5_script_pattern = re_compile(self.script_name + "-" + "[a-f0-9]{32}" + ".js")
        if (view_base / "_static").exists():
            self._bundled_scripts = [
                path.name
                for path in (view_base / "_static").iterdir()
                if md5_script_pattern.match(path.name) and ".js.map" not in path.name
            ]
            if not self._bundled_scripts:
                found_dependencies = False
            LOGGER.debug(
                f"[{self.__class__.__name__}] Resolved paths... {self._bundled_scripts}"
            )
        else:
            found_dependencies = False

        return found_dependencies

    @property
    def script_name(self):
        """
        The short-hand name of the controller, used to resolve the SSR script and other dependencies
        from disk.

        """
        return underscore(self.__class__.__name__)


def class_fn_as_method(fn):
    """
    Converts a class-bound action where `self` is not passed as the first argument
    to a method where `self` is passed as the first argument. This lets our dependency
    injection resolution work as normal without misinterpreting `self` as a query parameter.

    """

    class FunctionWrapper:
        pass

    setattr(FunctionWrapper, fn.__name__, fn)
    cls = FunctionWrapper()
    return getattr(cls, fn.__name__)


def function_is_action(name, func):
    try:
        metadata = get_function_metadata(func)
        if metadata.action_type in {
            FunctionActionType.PASSTHROUGH,
            FunctionActionType.SIDEEFFECT,
        }:
            yield name, func, metadata
    except AttributeError:
        return


def get_client_functions_cls(cls) -> Iterable[tuple[str, Callable, FunctionMetadata]]:
    """
    Gets the client functions defined on the class level, so only includes the
    functions that are implemented at this class level (versus MRO superclasses).

    """
    # Only look at functions directly defined in this class using __dict__
    for name, func in cls.__dict__.items():
        if isfunction(func):
            for name, func, metadata in function_is_action(name, func):
                yield name, class_fn_as_method(func), metadata



from datetime import date, datetime, time
from json import dumps as json_dumps
from types import NoneType
from typing import Any
from uuid import UUID

from fastapi import UploadFile

from mountaineer.client_builder.parser import (
    EnumWrapper,
    ModelWrapper,
    SelfReference,
)
from mountaineer.client_builder.types import (
    DictOf,
    ListOf,
    LiteralOf,
    Or,
    SetOf,
    TupleOf,
    TypeDefinition,
)
from mountaineer.logging import LOGGER


class InterfaceBase:
    """
    Each interface builder owns the responsibility of converting a specific type of object
    to a discrete block of Typescript code. This is the main location where the python
    typehints are formatted into Typescript type annotations.

    """

    @classmethod
    def _get_annotated_value(cls, value):
        """Convert a field type to TypeScript type."""
        if isinstance(value, ModelWrapper):
            return value.name.global_name
        elif isinstance(value, EnumWrapper):
            return value.name.global_name
        else:
            complex_value = cls._handle_complex_type(value, requires_complex=True)
            if complex_value:
                return complex_value
            if isinstance(value, SelfReference):
                return value.name
            primitive_value = cls._map_primitive_type_to_typescript(value)
            if primitive_value:
                return primitive_value
            LOGGER.warning(f"Unable to map value, falling back to generic: {value}")
            return "any"

    @classmethod
    def _map_primitive_type_to_typescript(cls, py_type: type) -> str | None:
        """Map Python types to TypeScript types"""
        type_map = {
            str: "string",
            int: "number",
            float: "number",
            bool: "boolean",
            # TODO: We should cast this to a Date internally
            datetime: "string",
            date: "string",
            time: "string",
            UploadFile: "Blob",
            UUID: "string",
            None: "null",
            NoneType: "null",
            Any: "any",
        }
        return type_map.get(py_type)

    @classmethod
    def _handle_complex_type(
        cls, type_hint: Any, requires_complex: bool = False
    ) -> str | None:
        """Handle complex type hints like list[str], dict[str, int], etc."""
        if not isinstance(type_hint, TypeDefinition):
            return None

        if isinstance(type_hint, ListOf):
            return f"Array<{cls._get_annotated_value(type_hint.type)}>"

        if isinstance(type_hint, TupleOf):
            values = [cls._get_annotated_value(t) for t in type_hint.types]
            return f"[{','.join(values)}]"

        if isinstance(type_hint, SetOf):
            return f"Set<{cls._get_annotated_value(type_hint.type)}>"

        if isinstance(type_hint, DictOf):
            return f"Record<{cls._get_annotated_value(type_hint.key_type)}, {cls._get_annotated_value(type_hint.value_type)}>"

        if isinstance(type_hint, Or):
            return " | ".join(cls._get_annotated_value(t) for t in type_hint.children)

        if isinstance(type_hint, LiteralOf):
            return " | ".join(json_dumps(value) for value in type_hint.values)

        raise ValueError(f"Unsupported TypeDefinition type: {type_hint}")



from typing import Annotated, Any, Dict, List, Literal, Optional, Union

import pytest

from mountaineer.client_builder.types import (
    DictOf,
    ListOf,
    LiteralOf,
    Or,
    SetOf,
    TupleOf,
    TypeDefinition,
    TypeParser,
    get_union_types,
    is_union_type,
)


class TypeComparisonHelpers:
    @staticmethod
    def are_types_equivalent(type1: Any, type2: Any) -> bool:
        """
        Recursively compares two type definitions for logical equivalence,
        handling nested types and different ordering of union types.
        """
        if isinstance(type1, Or) and isinstance(type2, Or):
            return TypeComparisonHelpers.are_or_types_equivalent(type1, type2)

        if isinstance(type1, LiteralOf) and isinstance(type2, LiteralOf):
            return set(type1.values) == set(type2.values)

        if isinstance(type1, TypeDefinition) and isinstance(type2, TypeDefinition):
            return TypeComparisonHelpers.are_type_definitions_equivalent(type1, type2)

        return bool(type1 == type2)

    @staticmethod
    def find_matching_type(target: Any, candidates: tuple[Any, ...]) -> bool:
        """
        Finds a matching type in candidates that is equivalent to the target.
        Handles unhashable types by doing direct comparisons.
        """
        return any(
            TypeComparisonHelpers.are_types_equivalent(target, candidate)
            for candidate in candidates
        )

    @staticmethod
    def are_or_types_equivalent(type1: Or, type2: Or) -> bool:
        """
        Compares two Or types for equivalence, handling different ordering of types.
        Uses list-based comparison instead of sets for unhashable types.
        """
        if len(type1.types) != len(type2.types):
            return False

        # Convert to lists to maintain order and handle unhashable types
        types1 = list(type1.types)
        types2 = list(type2.types)

        # For each type in types1, find and remove a matching type in types2
        remaining_types2 = list(types2)
        for t1 in types1:
            found_match = False
            for i, t2 in enumerate(remaining_types2):
                if TypeComparisonHelpers.are_types_equivalent(t1, t2):
                    remaining_types2.pop(i)
                    found_match = True
                    break
            if not found_match:
                return False

        return len(remaining_types2) == 0

    @staticmethod
    def are_type_definitions_equivalent(
        def1: TypeDefinition, def2: TypeDefinition
    ) -> bool:
        """
        Compares two TypeDefinition instances for equivalence by comparing their attributes.
        """
        if type(def1) is not type(def2):
            return False

        non_type_1 = [
            child for child in def1.children if not isinstance(child, TypeDefinition)
        ]
        non_type_2 = [
            child for child in def2.children if not isinstance(child, TypeDefinition)
        ]

        if non_type_1 != non_type_2:
            return False

        child_types_1 = [
            child for child in def1.children if isinstance(child, TypeDefinition)
        ]
        child_types_2 = [
            child for child in def2.children if isinstance(child, TypeDefinition)
        ]

        if len(child_types_1) != len(child_types_2):
            return False

        for a, b in zip(child_types_1, child_types_2):
            if not TypeComparisonHelpers.are_types_equivalent(a, b):
                return False

        return True


@pytest.fixture
def parser():
    return TypeParser()


@pytest.fixture
def type_compare():
    return TypeComparisonHelpers()


class TestUnionTypeDetection:
    @pytest.mark.parametrize(
        "type_input,expected",
        [
            (Union[str, int], True),
            (Optional[str], True),
            (str | int, True),
            (str | None, True),
            (list[int], False),
            (dict[str, int], False),
            (tuple[int, str], False),
            (set[float], False),
            (List[int], False),
            (int, False),
            (Any, False),
        ],
    )
    def test_is_union_type(self, type_input, expected):
        assert is_union_type(type_input) == expected

    @pytest.mark.parametrize(
        "type_input,expected",
        [
            (Union[str, int], [str, int]),
            (Optional[str], [str, type(None)]),
            (str | int, [str, int]),
            (str | None, [str, type(None)]),
            (int | str | float, [int, str, float]),
            (str | int | None, [str, int, type(None)]),
        ],
    )
    def test_get_union_types(self, type_input, expected):
        result = get_union_types(type_input)
        assert set(result) == set(expected)

    def test_get_union_types_invalid(self):
        with pytest.raises(ValueError):
            get_union_types(list[int])


class TestModernTypeSyntax:
    @pytest.mark.parametrize(
        "input_type,expected_type",
        [
            (str | int, Or(str, int)),
            (str | None, Or(str, type(None))),
            (int | str | float, Or(int, str, float)),
            (list[int], ListOf(int)),
            (dict[str, int], DictOf(str, int)),
            (tuple[str, int], TupleOf(str, int)),
            (set[int], SetOf(int)),
            (list[str | int], ListOf(Or(str, int))),
            (
                dict[str, list[int]],
                DictOf(str, ListOf(int)),
            ),
            (
                tuple[int, str | None],
                TupleOf(int, Or(str, type(None))),
            ),
            (
                dict[str | int, list[tuple[int, str]]],
                DictOf(Or(str, int), ListOf(TupleOf(int, str))),
            ),
            (
                list[dict[str, int | None]],
                ListOf(DictOf(str, Or(int, type(None)))),
            ),
        ],
    )
    def test_modern_type_syntax(self, parser, type_compare, input_type, expected_type):
        result = parser.parse_type(input_type)
        assert isinstance(result, type(expected_type))
        assert type_compare.are_types_equivalent(result, expected_type)

    def test_complex_modern_nested_types(self, parser, type_compare):
        complex_type = dict[str, list[tuple[int | None, str] | set[bool]]]
        result = parser.parse_type(complex_type)

        expected = DictOf(
            key=str,
            value=ListOf(
                Or(
                    TupleOf(Or(int, type(None)), str),
                    SetOf(bool),
                )
            ),
        )

        assert type_compare.are_types_equivalent(result, expected)

    @pytest.mark.parametrize(
        "input_type",
        [
            dict[str | int, list[set[bool] | tuple[int, ...]]],
            list[dict[str | None, set[int | str]]],
            tuple[list[int | None], dict[str, set[bool | None]]],
            set[tuple[int | None, str | bool]],
        ],
    )
    def test_various_modern_combinations(self, parser, input_type):
        result = parser.parse_type(input_type)
        assert isinstance(result, TypeDefinition)


class TestLiteralTypes:
    @pytest.mark.parametrize(
        "input_type,expected_type",
        [
            (Literal["a", "b"], LiteralOf("a", "b")),
            (Literal[1, 2, 3], LiteralOf(1, 2, 3)),
            (Literal[True, False], LiteralOf(True, False)),
            (Literal[None], LiteralOf(None)),
            (Literal["a", 1, True, None], LiteralOf("a", 1, True, None)),
            (list[Literal["a", "b"]], ListOf(LiteralOf("a", "b"))),
            (
                dict[str, Literal[1, 2, 3]],
                DictOf(str, LiteralOf(1, 2, 3)),
            ),
            (
                Literal["a", "b"] | int,
                Or(LiteralOf("a", "b"), int),
            ),
            (
                dict[Literal["x", "y"], list[Literal[1, 2] | str]],
                DictOf(LiteralOf("x", "y"), ListOf(type=Or(LiteralOf(1, 2), str))),
            ),
        ],
    )
    def test_literal_types(self, parser, type_compare, input_type, expected_type):
        result = parser.parse_type(input_type)
        assert isinstance(result, type(expected_type))
        assert type_compare.are_types_equivalent(result, expected_type)

    def test_invalid_literal_values(self, parser):
        with pytest.raises(TypeError):
            parser.parse_type(Literal[object()])

        with pytest.raises(TypeError):
            parser.parse_type(Literal[[1, 2, 3]])

        with pytest.raises(TypeError):
            parser.parse_type(Literal[{"a": 1}])


class TestAnnotatedTypes:
    """Tests for handling typing.Annotated types"""

    @pytest.mark.parametrize(
        "input_type,expected_type",
        [
            (Annotated[str, "metadata"], str),
            (Annotated[int, "metadata1", "metadata2"], int),
            (Annotated[List[int], "metadata"], ListOf(int)),
            (Annotated[str | int, "metadata"], Or(str, int)),
            (Annotated[Dict[str, int], "metadata"], DictOf(str, int)),
            (Annotated[str, "metadata"], str),
            # For nested Annotated types, the inner type is wrapped in Or
            (List[Annotated[str, "metadata"]], ListOf(Or(str))),
            (Dict[str, Annotated[int, "metadata"]], DictOf(str, Or(int))),
            (Annotated[str | None, "metadata"], Or(str, type(None))),
        ],
    )
    def test_annotated_types(self, parser, type_compare, input_type, expected_type):
        """Test that Annotated types are correctly parsed"""
        result = parser.parse_type(input_type)

        # For simple types like str, int, the result will be wrapped in Or
        if not isinstance(expected_type, TypeDefinition):
            expected_type = Or(expected_type)

        assert type_compare.are_types_equivalent(result, expected_type)

    def test_nested_annotated_types(self, parser, type_compare):
        """Test nested Annotated types"""
        # Complex nested type with Annotated
        complex_type = List[
            Annotated[Dict[str, Annotated[int | str, "metadata"]], "outer"]
        ]
        result = parser.parse_type(complex_type)

        # Expected structure after parsing
        expected = ListOf(DictOf(key=str, value=Or(int, str)))

        assert type_compare.are_types_equivalent(result, expected)

    def test_annotated_with_no_args(self, parser):
        """Test that Annotated with no args raises ValueError"""
        # We can't directly create an Annotated with no args in a type-safe way
        # So we'll test the error handling by mocking the situation

        # Create a mock field_type and origin_type for _parse_origin_type
        parser._parse_origin_type(Annotated[str, "metadata"], Annotated)

        with pytest.raises(ValueError):
            # Modify args to be empty to trigger the error
            parser._parse_origin_type(object(), Annotated)



from typing import Generic, TypeVar

from mountaineer.generics import expand_typevars, get_typevar_mapping

T = TypeVar("T")
K = TypeVar("K")


class Base(Generic[T]):
    pass


class Intermediate(Base[T], Generic[T, K]):
    pass


class Final(Intermediate[int, str]):
    pass


def test_get_typevar_mapping():
    mapping = get_typevar_mapping(Final)

    assert mapping == {
        T: int,
        K: str,
    }


def test_expand_typevars():
    assert expand_typevars(
        {
            K: T,
            T: int,
        }
    ) == {
        K: int,
        T: int,
    }



from enum import Enum
from typing import Type

from pydantic import BaseModel

from mountaineer.actions.fields import FunctionActionType
from mountaineer.client_builder.parser import (
    ActionWrapper,
    ControllerWrapper,
    EnumWrapper,
    ExceptionWrapper,
    FieldWrapper,
    ModelWrapper,
    WrapperName,
)
from mountaineer.client_builder.types import TypeDefinition
from mountaineer.controller import ControllerBase
from mountaineer.exceptions import APIException


def create_model_wrapper(
    model: Type[BaseModel],
    name: str,
    fields: list[FieldWrapper] | None = None,
    superclasses: list[ModelWrapper] | None = None,
) -> ModelWrapper:
    wrapper_name = WrapperName(name)
    return ModelWrapper(
        name=wrapper_name,
        module_name="test_module",
        model=model,
        isolated_model=model,  # Simplified for testing
        superclasses=superclasses or [],
        value_models=fields or [],
    )


# Helper function to create field wrappers
def create_field_wrapper(
    name: str,
    type_hint: type | ModelWrapper | EnumWrapper | TypeDefinition,
    required: bool = True,
) -> FieldWrapper:
    return FieldWrapper(name=name, value=type_hint, required=required)


# Helper function to create exception wrappers
def create_exception_wrapper(
    exception: Type[APIException],
    name: str,
    status_code: int,
    value_models: list[FieldWrapper] | None = None,
) -> ExceptionWrapper:
    wrapper_name = WrapperName(name)
    return ExceptionWrapper(
        name=wrapper_name,
        module_name="test_module",
        status_code=status_code,
        exception=exception,
        value_models=value_models or [],
    )


def create_action_wrapper(
    name: str,
    params: list[FieldWrapper] | None = None,
    response_model: Type[BaseModel] | None = None,
    request_body: ModelWrapper | None = None,
    action_type: FunctionActionType = FunctionActionType.PASSTHROUGH,
) -> ActionWrapper:
    response_wrapper = (
        create_model_wrapper(response_model, response_model.__name__)
        if response_model
        else None
    )
    return ActionWrapper(
        name=name,
        module_name="test_module",
        action_type=action_type,
        params=params or [],
        headers=[],
        request_body=request_body,
        response_bodies={ControllerBase: response_wrapper} if response_wrapper else {},
        exceptions=[],
        is_raw_response=False,
        is_streaming_response=False,
        controller_to_url={ControllerBase: f"/api/{name}"},
    )


def create_controller_wrapper(
    name: str,
    actions: dict[str, ActionWrapper] | None = None,
    superclasses: list[ControllerWrapper] | None = None,
    entrypoint_url: str | None = None,
) -> ControllerWrapper:
    wrapper_name = WrapperName(name)
    return ControllerWrapper(
        name=wrapper_name,
        module_name="test_module",
        entrypoint_url=entrypoint_url,
        controller=type(name, (ControllerBase,), {}),
        superclasses=superclasses or [],
        queries=[],
        paths=[],
        render=None,
        actions=actions or {},
    )


def create_enum_wrapper(enum_class: Type[Enum]) -> EnumWrapper:
    """Helper function to create enum wrappers"""
    wrapper_name = WrapperName(enum_class.__name__)
    return EnumWrapper(name=wrapper_name, module_name="test_module", enum=enum_class)



from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import AsyncIterator, Generic, Optional, TypeVar

import pytest
from fastapi import File, Form, UploadFile
from pydantic import BaseModel, ConfigDict, field_validator

from mountaineer.__tests__.client_builder.interface_builders.common import (
    create_field_wrapper,
    create_model_wrapper,
)
from mountaineer.actions.fields import FunctionActionType
from mountaineer.actions.passthrough_dec import passthrough
from mountaineer.actions.sideeffect_dec import sideeffect
from mountaineer.app import AppController
from mountaineer.client_builder.parser import (
    ActionWrapper,
    ControllerParser,
    ControllerWrapper,
    EnumWrapper,
    FieldWrapper,
    ModelWrapper,
)
from mountaineer.client_builder.types import ListOf
from mountaineer.controller import ControllerBase
from mountaineer.render import RenderBase


# Create a simple response model for testing
class StandardResponse(BaseModel):
    message: str


# Type variables for generic tests
T = TypeVar("T")
S = TypeVar("S")


# Core test enum
class ExampleEnum(Enum):
    A = "a"
    B = "b"
    C = "c"


# Core test models
class ExampleModelBase(BaseModel):
    string_field: str
    int_field: int
    enum_field: ExampleEnum

    @field_validator("string_field")
    def validate_string(cls, v):
        if len(v) < 3:
            raise ValueError("String too short")
        return v.upper()


# Generic test models
class GenericTestModel(BaseModel, Generic[T]):
    value: T
    metadata: str


class MultiGenericTestModel(GenericTestModel[T], Generic[T, S]):
    second_value: S


class NestedGenericTestModel(BaseModel, Generic[T]):
    wrapper: GenericTestModel[T]
    list_of: list[GenericTestModel[T]]


# Inheritance test models
class BaseInheritanceModel(ExampleModelBase):
    base_field: str


class LeftInheritanceModel(BaseInheritanceModel):
    left_field: int


class RightInheritanceModel(BaseInheritanceModel):
    right_field: float


class DiamondInheritanceModel(LeftInheritanceModel, RightInheritanceModel):
    final_field: bool


# Circular reference model
class CircularModel(BaseModel):
    name: str
    parent: Optional["CircularModel"] = None


CircularModel.model_rebuild()


# Controller response models
class ControllerResponse(BaseModel):
    message: str
    timestamp: datetime


class FileUploadResponse(BaseModel):
    filename: str
    size: int


class RenderResponse(RenderBase):
    data: ExampleModelBase
    count: int = 0


# Controller hierarchy
class BaseExampleController:
    @passthrough
    def base_action(self) -> ControllerResponse:  # type: ignore
        pass


class ExampleController(ControllerBase, BaseExampleController):
    url = "/test/{path_param}"
    view_path = "/test.tsx"

    async def render(self, path_param: str, query_param: int = 0) -> RenderResponse:  # type: ignore
        pass

    @passthrough
    def get_data(self) -> ExampleModelBase:  # type: ignore
        pass

    @sideeffect
    def update_form(  # type: ignore
        self, name: str = Form(...), size: int = Form(...)
    ) -> FileUploadResponse:  # type: ignore
        pass

    @sideeffect
    def upload_file(self, file: UploadFile = File(...)) -> FileUploadResponse:  # type: ignore
        pass


class SpecialTypesController(ControllerBase):
    url = "/test2"
    view_path = "/test2.tsx"

    @passthrough(raw_response=True)  # type: ignore
    async def raw_action(self) -> ExampleModelBase:  # type: ignore
        pass

    @passthrough
    async def stream_action(self) -> AsyncIterator[ExampleModelBase]:
        yield ExampleModelBase(
            string_field="test", int_field=1, enum_field=ExampleEnum.A
        )


# Tests
class TestControllerParser:
    @pytest.fixture
    def parser(self):
        parser = ControllerParser()
        app_controller = AppController(view_root=Path())
        app_controller.register(ExampleController())
        return parser

    @pytest.fixture
    def base_model_wrapper(self, parser: ControllerParser):
        return parser._parse_model(ExampleModelBase)

    @pytest.fixture
    def test_controller_wrapper(self, parser: ControllerParser):
        return parser.parse_controller(ExampleController)

    def test_parse_enum(self, parser: ControllerParser):
        wrapper = parser._parse_enum(ExampleEnum)
        assert isinstance(wrapper, EnumWrapper)
        assert wrapper.enum == ExampleEnum
        assert wrapper.name.raw_name == "ExampleEnum"

    def test_parse_base_model(self, parser: ControllerParser):
        wrapper = parser._parse_model(ExampleModelBase)
        assert isinstance(wrapper, ModelWrapper)
        assert wrapper.model == ExampleModelBase
        assert len(wrapper.value_models) == 3
        assert wrapper.superclasses == []

    def test_parse_generic_model(self, parser: ControllerParser):
        wrapper = parser._parse_model(GenericTestModel[str])
        assert isinstance(wrapper, ModelWrapper)
        assert wrapper.model == GenericTestModel[str]
        assert len(wrapper.value_models) == 2

    def test_parse_nested_generic_model(self, parser: ControllerParser):
        wrapper = parser._parse_model(NestedGenericTestModel[int])
        assert isinstance(wrapper, ModelWrapper)
        assert len(wrapper.value_models) == 2
        # Verify nested models were parsed
        assert any(isinstance(f.value, ModelWrapper) for f in wrapper.value_models)

    def test_parse_circular_model(self, parser: ControllerParser):
        wrapper = parser._parse_model(CircularModel)
        assert isinstance(wrapper, ModelWrapper)
        assert len(wrapper.value_models) == 2
        assert any(f.name == "parent" for f in wrapper.value_models)
        assert len(parser.parsed_self_references) == 1

    def test_parse_controller_inheritance(self, parser: ControllerParser):
        wrapper = parser.parse_controller(ExampleController)
        assert isinstance(wrapper, ControllerWrapper)
        assert len(wrapper.superclasses) == 1
        assert wrapper.superclasses[0].controller == BaseExampleController

    def test_parse_controller_actions(self, parser: ControllerParser):
        wrapper = parser.parse_controller(ExampleController)
        assert len(wrapper.actions) == 3

        # Actions explicitly tied to this class
        known_passthrough = {"get_data"}
        known_sideeffect = {"update_form", "upload_file"}

        # Verify different action types
        assert {
            a.name
            for a in wrapper.actions.values()
            if a.action_type == FunctionActionType.PASSTHROUGH
        } == known_passthrough
        assert {
            a.name
            for a in wrapper.actions.values()
            if a.action_type == FunctionActionType.SIDEEFFECT
        } == known_sideeffect

    def test_parse_render_method(self, parser: ControllerParser):
        wrapper = parser.parse_controller(ExampleController)
        assert wrapper.render is not None
        assert len(wrapper.paths) == 1
        assert len(wrapper.queries) == 1

    def test_parse_form_action(self, parser: ControllerParser):
        wrapper = parser.parse_controller(ExampleController)
        action = wrapper.actions.get("update_form")
        assert action is not None
        assert action.request_body is not None
        assert action.request_body.body_type == "application/x-www-form-urlencoded"

    def test_parse_file_upload_action(self, parser: ControllerParser):
        wrapper = parser.parse_controller(ExampleController)
        action = wrapper.actions.get("upload_file")
        assert action is not None
        assert action.request_body is not None
        assert action.request_body.body_type == "multipart/form-data"

    def test_parse_raw_response(self, parser: ControllerParser):
        wrapper = parser.parse_controller(SpecialTypesController)
        action = wrapper.actions.get("raw_action")
        assert action is not None
        assert action.is_raw_response

    def test_parse_server_side_renderer(self, parser: ControllerParser):
        wrapper = parser.parse_controller(SpecialTypesController)
        action = wrapper.actions.get("stream_action")
        assert action is not None
        assert action.is_streaming_response

    def test_parse_multiple_response_types(self, parser: ControllerParser):
        """
        Test that one action that's shared by multiple children will have
        a response type that includes all possible response models.

        """

        class ResponseA(RenderBase):
            pass

        class ResponseB(RenderBase):
            pass

        class MultiResponseParent(ControllerBase):
            @sideeffect
            def multi_action(self) -> None:
                pass

        class ResponseAController(MultiResponseParent):
            url = "/response_a"
            view_path = "/response_a.tsx"

            async def render(self) -> ResponseA:  # type: ignore
                pass

        class ResponseBController(MultiResponseParent):
            url = "/response_b"
            view_path = "/response_b.tsx"

            async def render(self) -> ResponseB:  # type: ignore
                pass

        app_controller = AppController(view_root=Path())
        app_controller.register(ResponseAController())
        app_controller.register(ResponseBController())

        wrapper: ControllerWrapper = parser.parse_controller(MultiResponseParent)
        action = wrapper.actions["multi_action"]

        a_response = action.response_bodies[ResponseAController]
        b_response = action.response_bodies[ResponseBController]

        assert a_response
        assert b_response

        response_a_sideeffect = next(
            field for field in a_response.value_models if field.name == "sideeffect"
        )
        response_b_sideeffect = next(
            field for field in b_response.value_models if field.name == "sideeffect"
        )

        assert isinstance(response_a_sideeffect.value, ModelWrapper)
        assert isinstance(response_b_sideeffect.value, ModelWrapper)

        assert response_a_sideeffect.value.model == ResponseA
        assert response_b_sideeffect.value.model == ResponseB


class TestInheritanceHandling:
    @pytest.fixture
    def parser(self):
        return ControllerParser()

    def test_basic_inheritance(self, parser: ControllerParser):
        wrapper = parser._parse_model(BaseInheritanceModel)
        assert len(wrapper.superclasses) == 1
        assert wrapper.superclasses[0].model == ExampleModelBase

    def test_diamond_inheritance(self, parser: ControllerParser):
        wrapper = parser._parse_model(DiamondInheritanceModel)
        assert len([sc.name.global_name for sc in wrapper.superclasses]) == 2
        superclass_models = {s.model for s in wrapper.superclasses}
        assert LeftInheritanceModel in superclass_models
        assert RightInheritanceModel in superclass_models


class TestGenericHandling:
    @pytest.fixture
    def parser(self):
        return ControllerParser()

    def test_basic_generic(self, parser: ControllerParser):
        wrapper = parser._parse_model(GenericTestModel[str])
        assert len(wrapper.value_models) == 2
        assert wrapper.value_models[0].value is str

    def test_multi_generic(self, parser: ControllerParser):
        wrapper = parser._parse_model(MultiGenericTestModel[str, int])
        assert len(wrapper.value_models) == 1
        assert any(
            f.name == "second_value" and f.value is int for f in wrapper.value_models
        )

    def test_nested_generic_resolution(self, parser: ControllerParser):
        wrapper = parser._parse_model(NestedGenericTestModel[str])
        assert len(wrapper.value_models) == 2
        list_field = next(f for f in wrapper.value_models if f.name == "list_of")
        assert isinstance(list_field.value, ListOf)
        assert isinstance(list_field.value.children[0], ModelWrapper)
        assert "GenericTestModel[str]" in str(list_field.value.children[0].model)


class TestControllerWrapperFeatures:
    @pytest.fixture
    def parser(self):
        parser = ControllerParser()
        app_controller = AppController(view_root=Path())
        app_controller.register(ExampleController())
        return parser

    def test_all_actions_collection(self, parser: ControllerParser):
        wrapper = parser.parse_controller(ExampleController)
        action_names = {action.name for action in wrapper.all_actions}
        assert action_names == {"base_action", "get_data", "update_form", "upload_file"}

    @pytest.mark.parametrize(
        "include_superclasses, expected_models",
        [
            (False, {"ExampleModelBase", "FileUploadResponse", "RenderResponse"}),
            (
                True,
                {
                    "ExampleModelBase",
                    "ControllerResponse",
                    "FileUploadResponse",
                    "RenderResponse",
                },
            ),
        ],
    )
    def test_embedded_types_collection(
        self,
        parser: ControllerParser,
        include_superclasses: bool,
        expected_models: set[str],
    ):
        wrapper = parser.parse_controller(ExampleController)
        embedded = ControllerWrapper.get_all_embedded_types(
            [wrapper], include_superclasses=include_superclasses
        )

        model_names = {m.model.__name__ for m in embedded.models}
        assert model_names.intersection(expected_models) == expected_models

    def test_embedded_controllers(self, parser: ControllerParser):
        wrapper = parser.parse_controller(ExampleController)
        controllers = ControllerWrapper.get_all_embedded_controllers([wrapper])

        controller_types = {c.controller for c in controllers}
        assert controller_types == {ExampleController, BaseExampleController}


class TestIsolatedModelCreation:
    @pytest.fixture
    def parser(self):
        return ControllerParser()

    def test_standard_model_isolation(self, parser: ControllerParser):
        class ParentModel(BaseModel):
            parent_field: str
            shared_field: int = 0

        class ChildModel(ParentModel):
            child_field: str
            shared_field: int = 1  # Override parent field

        isolated = parser._create_isolated_model(ChildModel)

        # Should only include fields directly defined in ChildModel
        assert set(isolated.model_fields.keys()) == {"child_field", "shared_field"}
        assert "parent_field" not in isolated.model_fields

        # Check that field types are preserved
        assert isolated.model_fields["child_field"].annotation is str
        assert isolated.model_fields["shared_field"].annotation is int

    def test_nested_inheritance_isolation(self, parser: ControllerParser):
        class GrandparentModel(BaseModel):
            grandparent_field: str

        class ParentModel(GrandparentModel):
            parent_field: int

        class ChildModel(ParentModel):
            child_field: bool

        isolated = parser._create_isolated_model(ChildModel)

        # Should only include fields from ChildModel
        assert set(isolated.model_fields.keys()) == {"child_field"}
        assert "parent_field" not in isolated.model_fields
        assert "grandparent_field" not in isolated.model_fields

    def test_generic_model_isolation(self, parser: ControllerParser):
        class GenericParent(BaseModel, Generic[T]):
            parent_field: T
            shared_field: str = "parent"

        class GenericChild(GenericParent[int]):
            child_field: str
            shared_field: str = "child"  # Override parent field

        isolated = parser._create_isolated_model(GenericChild)

        # Should only include fields defined directly in GenericChild
        assert set(isolated.model_fields.keys()) == {"child_field", "shared_field"}
        assert "parent_field" not in isolated.model_fields

        # Verify field types are correctly resolved
        assert isolated.model_fields["child_field"].annotation is str
        assert isolated.model_fields["shared_field"].annotation is str

    def test_multi_generic_model_isolation(self, parser: ControllerParser):
        class MultiGenericParent(BaseModel, Generic[T, S]):
            field_t: T
            field_s: S

        class MultiGenericChild(MultiGenericParent[str, int]):
            child_field: bool

        isolated = parser._create_isolated_model(MultiGenericChild)

        # Should only include fields defined in child
        assert set(isolated.model_fields.keys()) == {"child_field"}
        assert "field_t" not in isolated.model_fields
        assert "field_s" not in isolated.model_fields

        # Verify field type
        assert isolated.model_fields["child_field"].annotation is bool

    def test_model_config_preservation(self, parser: ControllerParser):
        class CustomModel(BaseModel):
            field: str

            model_config = ConfigDict(
                str_strip_whitespace=True,
                frozen=True,
            )

        isolated = parser._create_isolated_model(CustomModel)

        # Check that model configuration is preserved
        assert isolated.model_config["str_strip_whitespace"] is True  # type: ignore
        assert isolated.model_config["frozen"] is True  # type: ignore

    def test_empty_model_isolation(self, parser: ControllerParser):
        class EmptyParent(BaseModel):
            parent_field: str

        class EmptyChild(EmptyParent):
            pass  # No direct fields

        isolated = parser._create_isolated_model(EmptyChild)

        # Should have no fields since child defines none directly
        assert not isolated.model_fields


class TestActionWrapper:
    @pytest.fixture
    def parser(self):
        parser = ControllerParser()
        app_controller = AppController(view_root=Path())
        app_controller.register(ExampleController())
        return parser

    @pytest.mark.parametrize(
        "params, headers, request_body, expected_has_required",
        [
            # Base cases - no headers or request body
            ([], [], None, False),
            ([create_field_wrapper("optional_param", str, False)], [], None, False),
            ([create_field_wrapper("required_param", str, True)], [], None, True),
            # Headers only
            ([], [create_field_wrapper("optional_header", str, False)], None, False),
            ([], [create_field_wrapper("required_header", str, True)], None, True),
            (
                [],
                [
                    create_field_wrapper("required_header", str, True),
                    create_field_wrapper("optional_header", str, False),
                ],
                None,
                True,
            ),
            # Request body only
            (
                [],
                [],
                create_model_wrapper(ExampleModelBase, "ExampleModelBase"),
                True,  # Request body is always required
            ),
            # Combination of params, headers, and request body
            (
                [create_field_wrapper("optional_param", str, False)],
                [create_field_wrapper("optional_header", str, False)],
                None,
                False,
            ),
            (
                [create_field_wrapper("required_param", str, True)],
                [create_field_wrapper("optional_header", str, False)],
                None,
                True,
            ),
            (
                [create_field_wrapper("optional_param", str, False)],
                [create_field_wrapper("required_header", str, True)],
                None,
                True,
            ),
            (
                [create_field_wrapper("optional_param", str, False)],
                [create_field_wrapper("optional_header", str, False)],
                create_model_wrapper(ExampleModelBase, "ExampleModelBase"),
                True,  # Request body makes it required
            ),
            (
                [create_field_wrapper("required_param", str, True)],
                [create_field_wrapper("required_header", str, True)],
                create_model_wrapper(ExampleModelBase, "ExampleModelBase"),
                True,
            ),
        ],
    )
    def test_has_required_params(
        self,
        params: list[FieldWrapper],
        headers: list[FieldWrapper],
        request_body: Optional[ModelWrapper],
        expected_has_required: bool,
    ):
        action = ActionWrapper(
            name="test_action",
            module_name="test_module",
            action_type=FunctionActionType.PASSTHROUGH,
            params=params,
            headers=headers,
            request_body=request_body,
            response_bodies={
                ControllerBase: create_model_wrapper(
                    StandardResponse, "StandardResponse"
                )
            },
            exceptions=[],
            is_raw_response=False,
            is_streaming_response=False,
            controller_to_url={ControllerBase: "/api/test"},
        )

        assert action.has_required_params() == expected_has_required



from dataclasses import dataclass
from typing import Any

from mountaineer.client_builder.interface_builders.action import ActionInterface
from mountaineer.client_builder.interface_builders.base import InterfaceBase
from mountaineer.client_builder.parser import (
    ControllerWrapper,
)
from mountaineer.client_builder.typescript import (
    TSLiteral,
    python_payload_to_typescript,
)


@dataclass
class ControllerInterface(InterfaceBase):
    """
    The Controller interfaces consolidate all actions that belong to
    this controller in the backend. Render function inputs are stored
    separately in their respective ModelInterface.

    """

    name: str
    body: str
    include_superclasses: list[str]
    include_export: bool = True

    @classmethod
    def from_controller(cls, controller: ControllerWrapper):
        fields: dict[str, Any] = {}

        # Convert each action that's directly owned by the controller
        for url, action in controller.actions.items():
            # We don't need a URL here because we just want the type definition, not
            # the full definition
            action_def = ActionInterface.from_action(action, url="", controller=None)

            # If we only have optional params, we will default the inputs to an empty object
            # which makes the whole params object optional
            optional_params = "?" if action_def.default_initializer else ""

            if not action_def.response_type:
                raise ValueError(
                    f"Action {action_def.name} is missing an auto-detected response type"
                )

            action_signature = f"(params{optional_params}: {action_def.typehints}) => {action_def.response_type}"

            fields[TSLiteral(action_def.name)] = TSLiteral(action_signature)

        return cls(
            name=controller.name.global_name,
            body=python_payload_to_typescript(fields),
            include_superclasses=[s.name.global_name for s in controller.superclasses],
        )

    def to_js(self):
        schema_def = f"interface {self.name}"

        if self.include_superclasses:
            schema_def += f" extends {', '.join(self.include_superclasses)}"

        # Our stored body will include the starting/closing brackets
        schema_def += f" {self.body}"

        if self.include_export:
            schema_def = f"export {schema_def}"

        return schema_def



import asyncio
from pathlib import Path

import pytest
from watchfiles import Change

from mountaineer.development.watch import (
    CallbackDefinition,
    CallbackEvent,
    CallbackMetadata,
    CallbackType,
    FileWatcher,
    PackageWatchdog,
)


@pytest.mark.parametrize(
    "path, ignore_list, ignore_hidden, expected_ignore",
    [
        ("myproject/regular.py", ["__pycache__"], True, False),
        ("myproject/.hidden", [""], True, True),
        ("myproject/.hidden/subfile", [""], True, True),
        ("myproject/.hidden/subfile", [""], False, False),
        ("myproject/__cache__", ["__cache__"], True, True),
    ],
)
def test_ignore_path(
    path: str,
    ignore_list: list[str],
    ignore_hidden: bool,
    expected_ignore: bool,
):
    watcher = FileWatcher([], ignore_list=ignore_list, ignore_hidden=ignore_hidden)
    assert watcher.should_ignore_path(Path(path)) == expected_ignore


@pytest.mark.parametrize(
    "paths, expected_paths",
    [
        (
            # Unique paths
            ["myproject/regular.py", "myproject/.hidden"],
            ["myproject/regular.py", "myproject/.hidden"],
        ),
        (
            # Simple subdirectory
            ["myproject", "myproject/subdir"],
            ["myproject"],
        ),
        (
            # More specific directory first
            ["myproject/subdir1/subdir2", "myproject/subdir1"],
            ["myproject/subdir1"],
        ),
    ],
)
def test_merge_paths(paths: list[str], expected_paths: list[str]):
    handler = PackageWatchdog("mountaineer", [])
    assert set(handler.merge_paths(paths)) == {
        str(Path(path).absolute()) for path in expected_paths
    }


def test_change_mapping():
    watcher = FileWatcher([])
    assert watcher._map_change_to_callback_type(Change.added) == CallbackType.CREATED
    assert (
        watcher._map_change_to_callback_type(Change.modified) == CallbackType.MODIFIED
    )
    assert watcher._map_change_to_callback_type(Change.deleted) == CallbackType.DELETED


@pytest.mark.asyncio
async def test_file_notification(tmp_path: Path):
    callback_events: list[CallbackEvent] = []
    stop_event = asyncio.Event()

    async def receive_callback(metadata: CallbackMetadata):
        callback_events.extend(metadata.events)

    async def simulate_changes(watcher: FileWatcher):
        # Make sleeps longer than the debounce interval (0.1s)
        await asyncio.sleep(0.15)
        await watcher.process_changes([(Change.added, str(tmp_path / "test.txt"))])
        await asyncio.sleep(0.15)
        await watcher.process_changes([(Change.modified, str(tmp_path / "test.txt"))])
        await asyncio.sleep(0.15)
        await watcher.process_changes([(Change.deleted, str(tmp_path / "test.txt"))])
        await asyncio.sleep(0.15)
        stop_event.set()

    watcher = FileWatcher(
        callbacks=[
            CallbackDefinition(
                action=CallbackType.CREATED
                | CallbackType.MODIFIED
                | CallbackType.DELETED,
                callback=receive_callback,
            )
        ],
    )

    changes_task = asyncio.create_task(simulate_changes(watcher))
    try:
        await asyncio.wait_for(stop_event.wait(), timeout=2.0)
    finally:
        changes_task.cancel()

    event_actions: list[CallbackType] = []
    for event in callback_events:
        if event.action not in event_actions:
            event_actions.append(event.action)

    assert event_actions == [
        CallbackType.CREATED,
        CallbackType.MODIFIED,
        CallbackType.DELETED,
    ]



from dataclasses import dataclass
from typing import Any

from mountaineer.client_builder.interface_builders.base import InterfaceBase
from mountaineer.client_builder.parser import (
    ModelWrapper,
)
from mountaineer.client_builder.typescript import (
    TSLiteral,
    python_payload_to_typescript,
)


@dataclass
class ModelInterface(InterfaceBase):
    name: str
    body: str
    include_superclasses: list[str]
    include_export: bool = True

    @classmethod
    def from_model(cls, value: ModelWrapper):
        fields: dict[str, Any] = {}
        for field in value.value_models:
            field_name = f"{field.name}{'?' if not field.required else ''}"
            field_type = cls._get_annotated_value(field.value)
            fields[TSLiteral(field_name)] = TSLiteral(field_type)

        return cls(
            name=value.name.global_name,
            body=python_payload_to_typescript(fields),
            include_superclasses=[s.name.global_name for s in value.superclasses],
        )

    def to_js(self) -> str:
        schema_def = f"interface {self.name}"

        if self.include_superclasses:
            schema_def += f" extends {', '.join(self.include_superclasses)}"

        # Our stored body will include the starting/closing brackets
        schema_def += f" {self.body}"

        if self.include_export:
            schema_def = f"export {schema_def}"

        return schema_def



import logging
from json import dumps as json_dumps
from pathlib import Path
from re import finditer as re_finditer
from typing import Any, cast

from mountaineer import mountaineer as mountaineer_rs  # type: ignore
from mountaineer.cache import extended_lru_cache
from mountaineer.client_compiler.source_maps import SourceMapParser
from mountaineer.logging import debug_log_artifact
from mountaineer.static import get_static_path


class V8RuntimeError(Exception):
    """
    An exception thrown by the V8 runtime in the case of a permanent failure that
    involves the content of the script.

    """

    pass


def fix_exception_lines(*, exception: str, injected_script: str):
    """
    Since we create a synthetic script to run in the V8 runtime, the line numbers
    of the in-application stack trace will be offset by however long our
    injected script is. This function re-processed the stack trace to fix the
    line number.

    """
    offset_lines = injected_script.count("\n")

    text_replacements: dict[tuple[int, int], str] = {}
    for match in re_finditer(r"\(([<>A-Za-z0-9]+?):(\d+?):(\d+?)\)", exception):
        # Only replace the line numbers
        text_replacements[match.span(2)] = str(int(match.group(2)) - offset_lines)

    sorted_replacements = sorted(
        text_replacements.items(), key=lambda x: x[0][0], reverse=True
    )

    for (start, end), replacement in sorted_replacements:
        exception = exception[:start] + replacement + exception[end:]

    return exception


def find_tsconfig(paths: list[list[str]]) -> str | None:
    """
    Find the tsconfig.json file in the parent directories of the provided paths.
    We look for tsconfig.json in the parent directories of each path in the list.
    If multiple tsconfig.json files are found, we use the one closest to the root.

    :param paths: List of lists of file paths to search from
    :return: Path to tsconfig.json if found, None otherwise
    """
    tsconfig_paths = set()

    # For each group of paths
    for path_group in paths:
        # For each path in the group
        for path in path_group:
            current = Path(path).parent
            # Walk up the directory tree
            while current != current.parent:
                tsconfig = current / "tsconfig.json"
                if tsconfig.exists():
                    tsconfig_paths.add(str(tsconfig))
                current = current.parent

    if not tsconfig_paths:
        logging.warning(
            f"No tsconfig.json found in any parent directory of the provided paths: {paths}"
        )
        return None

    # Return the tsconfig.json closest to the original file
    return min(tsconfig_paths, key=len)


@extended_lru_cache(maxsize=128, max_size_mb=5)
def render_ssr(
    script: str,
    render_data: dict[str, Any],
    hard_timeout: int | float | None = None,
    sourcemap: str | None = None,
) -> str:
    """
    Render the React component in the provided SSR javascript bundle. This file will
    be directly executed within the V8 runtime.

    To speed up requests for the same exact content (ie. same react and same data)
    we cache the result of the render_ssr_rust call by default for a limited amount of
    previous calls. We limit the overall size of this cache to 5MB.

    :param script: The raw code of the javascript bundle to execute. Should be pre-compiled into an
        SSR compatible package with a single entrypoint.
    :param render_data: The data to inject into the SSR javascript bundle
    :param hard_timeout: The maximum time to allow the render to take in seconds. If the render takes
        longer than this time, our thread supervisor will kick in and terminate the rust worker.

    :raises TimeoutError: If the render takes longer than the hard_timeout
    :raises V8RuntimeError: If the V8 runtime throws an exception during the render

    """
    polyfill_script = get_static_path("ssr_polyfills.js").read_text()
    data_json = json_dumps(render_data)

    injected_script = f"var SERVER_DATA = {data_json};\n{polyfill_script}\n"
    full_script = f"{injected_script}{script}"

    debug_log_artifact("ssr_script", "js", full_script)

    try:
        # Convert to milliseconds for the rust worker
        render_result = mountaineer_rs.render_ssr(
            full_script, int(hard_timeout * 1000) if hard_timeout else 0
        )
    except ConnectionAbortedError:
        raise TimeoutError("SSR render was interrupted after hard timeout")
    except ValueError as e:
        js_stack = fix_exception_lines(
            exception=str(e), injected_script=injected_script
        )

        if sourcemap:
            sourcemap_parser = SourceMapParser(script=sourcemap)
            sourcemap_parser.parse()
            js_stack = sourcemap_parser.map_exception(js_stack)

        raise V8RuntimeError(js_stack)

    return cast(str, render_result)



from dataclasses import dataclass
from typing import Any, Type

from mountaineer.client_builder.file_generators.base import CodeBlock
from mountaineer.client_builder.interface_builders.base import InterfaceBase
from mountaineer.client_builder.parser import (
    ActionWrapper,
)
from mountaineer.client_builder.typescript import (
    TSLiteral,
    python_payload_to_typescript,
)
from mountaineer.controller import ControllerBase


@dataclass
class ActionInterface(InterfaceBase):
    name: str
    parameters: str
    typehints: str
    default_initializer: bool
    response_type: str
    body: list[str]

    def to_js(self) -> str:
        script = f"export const {self.name} = ({self.parameters} : {self.typehints}"
        if self.default_initializer:
            script += " = {}"
        body_str = "\n".join(self.body)
        script += f"): {self.response_type} => {{ {body_str} }}"
        return script

    @classmethod
    def from_action(
        cls, action: ActionWrapper, url: str, controller: Type[ControllerBase] | None
    ):
        """
        If controller is None, we should take the union of all response bodies. This is used for global definitions
        where the exact model might be different by the concrete controller (like for @sideeffects that must include
        the specific controller's render model).

        """
        parameters_dict: dict[str, Any] = {}
        typehint_dict: dict[str, Any] = {}

        # System parameters (always optional)
        system_parameters = {"signal": TSLiteral("signal")}
        system_typehints: dict[str, Any] = {
            TSLiteral("signal?"): TSLiteral("AbortSignal")
        }

        # Add path/query parameters
        for param in action.params:
            parameters_dict[param.name] = TSLiteral(param.name)
            typehint_dict[
                TSLiteral(f"{param.name}{'?' if not param.required else ''}")
            ] = TSLiteral(cls._get_annotated_value(param.value))

        for header in action.headers:
            parameters_dict[header.name] = TSLiteral(header.name)
            typehint_dict[
                TSLiteral(f"{header.name}{'?' if not header.required else ''}")
            ] = TSLiteral(cls._get_annotated_value(header.value))

        # Add request body if present
        if action.request_body:
            model_name = action.request_body.name.global_name
            parameters_dict["requestBody"] = TSLiteral("requestBody")
            typehint_dict[TSLiteral("requestBody")] = TSLiteral(model_name)

        # Merge system parameters
        parameters_dict.update(system_parameters)
        typehint_dict.update(system_typehints)

        request_payload = cls._build_request_payload(url, action, parameters_dict)

        response_types: set[str] = set()
        controllers = [controller] if controller else action.response_bodies.keys()
        if controllers:
            for controller in controllers:
                response_types.add(cls._get_response_type(action, controller))
        else:
            # Fallback in the case that no concrete controllers are mounted with this action
            # In this case we just use a generic typehint for the return value
            response_types.add(cls._get_response_type(action, None))

        return cls(
            name=action.name,
            parameters=python_payload_to_typescript(parameters_dict),
            typehints=python_payload_to_typescript(typehint_dict),
            default_initializer=not action.has_required_params(),
            response_type=" | ".join(response_types),
            body=["return __request(", CodeBlock.indent(f"  {request_payload}"), ");"],
        )

    @classmethod
    def _build_request_payload(
        cls, url: str, action: ActionWrapper, parameters: dict[str, Any]
    ) -> str:
        payload: dict[str, Any] = {
            "method": "POST",
            "url": url,
            "path": {},
            "headers": {},
            "query": {},
            "errors": {},
            "signal": TSLiteral("signal"),
        }

        for param in action.params:
            if param.name in parameters:
                payload["query"][param.name] = TSLiteral(param.name)

        for header in action.headers:
            if header.name in parameters:
                payload["headers"][header.name] = TSLiteral(header.name)

        if action.request_body:
            payload["body"] = TSLiteral("requestBody")
            payload["mediaType"] = action.request_body.body_type

        for exception in action.exceptions:
            payload["errors"][exception.status_code] = TSLiteral(
                exception.name.global_name
            )

        if action.is_raw_response:
            payload["outputFormat"] = "raw"

        # Support for server-events
        if action.is_streaming_response:
            payload["eventStreamResponse"] = True

        # Clean up empty dicts. These are optional in the request API interface and
        # clean up the generated code
        for key in ["path", "headers", "query", "errors"]:
            if not payload[key]:
                del payload[key]

        return python_payload_to_typescript(
            {TSLiteral(key): value for key, value in payload.items()}
        )

    @classmethod
    def _get_response_type(
        cls, action: ActionWrapper, controller: Type[ControllerBase] | None
    ) -> str:
        if action.is_raw_response:
            return "Promise<Response>"

        if not controller:
            return "Promise<any>"

        response_body = cls._get_response_body(action, controller)
        if not response_body:
            return "Promise<void>"

        response_type = response_body.name.global_name

        if action.is_streaming_response:
            return f"Promise<AsyncGenerator<{response_type}, void, unknown>>"

        return f"Promise<{response_type}>"

    @classmethod
    def _get_response_body(
        cls, action: ActionWrapper, controller: Type[ControllerBase]
    ):
        return action.response_bodies.get(controller)



from click import command, option, group
from mountaineer.cli import handle_runserver, handle_watch, handle_build
from mountaineer.io import async_to_sync
from mountaineer.dependencies import get_function_dependencies
from mountaineer import Depends, CoreDependencies

from iceaxe import DBConnection
from iceaxe.mountaineer import DatabaseConfig, DatabaseDependencies
from iceaxe.migrations.cli import handle_apply, handle_generate, handle_rollback
from iceaxe.schemas.cli import create_all

from {{project_name}} import models
from {{project_name}}.config import AppConfig


@command()
@option("--port", default=5006, help="Port to run the server on")
def runserver(port: int):
    handle_runserver(
        package="{{project_name}}",
        webservice="{{project_name}}.main:app",
        webcontroller="{{project_name}}.app:controller",
        port=port,
    )


@command()
def watch():
    handle_watch(
        package="{{project_name}}",
        webcontroller="{{project_name}}.app:controller",
    )


@command()
def build():
    handle_build(
        webcontroller="{{project_name}}.app:controller",
    )


@command()
@async_to_sync
async def createdb():
    _ = AppConfig()  # type: ignore

    async def run_bootstrap(
        db_connection: DBConnection = Depends(DatabaseDependencies.get_db_connection),
    ):
        await create_all(db_connection=db_connection)

    async with get_function_dependencies(callable=run_bootstrap) as values:
        await run_bootstrap(**values)


@group
def migrate():
    pass


@migrate.command()
@option("--message", required=False)
@async_to_sync
async def generate(message: str | None):
    async def _inner(
        config: AppConfig = Depends(CoreDependencies.get_config_with_type(AppConfig)),
        db_connection: DBConnection = Depends(DatabaseDependencies.get_db_connection),
    ):
        assert config.PACKAGE
        await handle_generate(config.PACKAGE, db_connection, message=message)

    _ = AppConfig()  # type: ignore
    async with get_function_dependencies(callable=_inner) as deps:
        await _inner(**deps)


@migrate.command()
@async_to_sync
async def apply():
    async def _inner(
        config: AppConfig = Depends(CoreDependencies.get_config_with_type(AppConfig)),
        db_connection: DBConnection = Depends(DatabaseDependencies.get_db_connection),
    ):
        assert config.PACKAGE
        await handle_apply(config.PACKAGE, db_connection)

    _ = AppConfig()  # type: ignore
    async with get_function_dependencies(callable=_inner) as deps:
        await _inner(**deps)


@migrate.command()
@async_to_sync
async def rollback():
    async def _inner(
        config: AppConfig = Depends(CoreDependencies.get_config_with_type(AppConfig)),
        db_connection: DBConnection = Depends(DatabaseDependencies.get_db_connection),
    ):
        assert config.PACKAGE
        await handle_rollback(config.PACKAGE, db_connection)

    _ = AppConfig()  # type: ignore
    async with get_function_dependencies(callable=_inner) as deps:
        await _inner(**deps)



from typing import Any

from mountaineer.actions.fields import FunctionActionType
from mountaineer.client_builder.file_generators.base import CodeBlock, FileGeneratorBase
from mountaineer.client_builder.interface_builders.action import ActionInterface
from mountaineer.client_builder.interface_builders.controller import ControllerInterface
from mountaineer.client_builder.parser import (
    ControllerWrapper,
    EnumWrapper,
    ExceptionWrapper,
    FieldWrapper,
    TypeDefinition,
)
from mountaineer.client_builder.typescript import (
    TSLiteral,
    python_payload_to_typescript,
)
from mountaineer.paths import ManagedViewPath, generate_relative_import


class LocalGeneratorBase(FileGeneratorBase):
    def __init__(self, *, managed_path: ManagedViewPath, global_root: ManagedViewPath):
        super().__init__(managed_path=managed_path)
        self.global_root = global_root

    def get_global_import_path(self, global_name: str):
        """
        Gets the relative path from the current file to the desired global file. This
        allows for typescript imports in the same workflow.

        """
        root_common_handler = self.global_root / global_name
        return generate_relative_import(self.managed_path, root_common_handler)


class LocalLinkGenerator(LocalGeneratorBase):
    def __init__(
        self,
        controller: ControllerWrapper,
        *,
        managed_path: ManagedViewPath,
        global_root: ManagedViewPath,
    ):
        super().__init__(managed_path=managed_path, global_root=global_root)
        self.controller = controller

    def script(self):
        yield from self._get_imports(self.controller)
        yield CodeBlock(self._get_link_implementation(self.controller))

    def _get_imports(self, controller: ControllerWrapper):
        api_import_path = self.get_global_import_path("api.ts")
        controller_import_path = self.get_global_import_path("controllers.ts")

        yield CodeBlock(f"import {{ __getLink }} from '{api_import_path}';")

        # Deeply traverse for any enums contained within these fields - this especially
        # is useful for nested objects, like an optional URL parameter (Or(None, Enum))
        imports = []

        def _traverse_logic(item: FieldWrapper | EnumWrapper | TypeDefinition):
            if isinstance(item, EnumWrapper):
                imports.append(
                    f"import {{ {item.name.global_name} }} from '{controller_import_path}';"
                )
            elif isinstance(item, TypeDefinition):
                yield from item.children
            elif isinstance(item, FieldWrapper):
                yield item.value

        ControllerWrapper._traverse_iterator(
            _traverse_logic, controller.queries + controller.paths
        )

        yield CodeBlock(*imports)

    def _get_link_implementation(self, controller: ControllerWrapper) -> str:
        """Generate link formatter for a controller's routes"""
        if not controller.render:
            return ""
        if not controller.entrypoint_url:
            return ""

        # Collect parameters from render model
        query_parameters: dict[str, Any] = {}
        path_parameters: dict[str, Any] = {}
        query_typehints: dict[str, Any] = {}
        path_typehints: dict[str, Any] = {}

        # Split parameters into query and path
        for field in controller.queries:
            query_parameters[field.name] = TSLiteral(field.name)
            optional_param = "?" if not field.required else ""
            query_typehints[TSLiteral(field.name + optional_param)] = TSLiteral(
                ControllerInterface._get_annotated_value(field.value)
            )

        for field in controller.paths:
            path_parameters[field.name] = TSLiteral(field.name)
            optional_param = "?" if not field.required else ""
            path_typehints[TSLiteral(field.name + optional_param)] = TSLiteral(
                ControllerInterface._get_annotated_value(field.value)
            )

        # Combine all parameters for the function signature
        all_parameters = {**query_parameters, **path_parameters}
        all_typehints = {**query_typehints, **path_typehints}

        return self._generate_link_function(
            controller.entrypoint_url,
            all_parameters,
            all_typehints,
            query_parameters,
            path_parameters,
        )

    def _generate_link_function(
        self,
        url: str,
        parameters: dict[str, Any],
        typehints: dict[str, Any],
        query_parameters: dict[str, Any],
        path_parameters: dict[str, Any],
    ) -> str:
        """Generate the TypeScript link function"""
        param_str = python_payload_to_typescript(parameters)
        typehint_str = python_payload_to_typescript(typehints)

        query_dict_str = python_payload_to_typescript(query_parameters)
        path_dict_str = python_payload_to_typescript(path_parameters)

        link_args = {
            TSLiteral("rawUrl"): TSLiteral("url"),
            TSLiteral("queryParameters"): TSLiteral("queryParameters"),
            TSLiteral("pathParameters"): TSLiteral("pathParameters"),
        }
        link_args_str = python_payload_to_typescript(link_args)

        link_logic = [
            f"const url = `{url}`;\n",
            f"const queryParameters: Record<string, any> = {query_dict_str};",
            f"const pathParameters: Record<string, any> = {path_dict_str};\n",
            CodeBlock.indent(f"return __getLink({link_args_str});"),
        ]
        link_logic_str = "\n".join(link_logic)

        # Function signature with all parameters
        lines = [
            f"export const getLink = ({param_str}: {typehint_str}) => {{",
            CodeBlock.indent(f"  {link_logic_str}"),
            "};",
        ]

        return "\n".join(lines)


class LocalActionGenerator(LocalGeneratorBase):
    def __init__(
        self,
        controller: ControllerWrapper,
        *,
        managed_path: ManagedViewPath,
        global_root: ManagedViewPath,
    ):
        super().__init__(managed_path=managed_path, global_root=global_root)
        self.controller = controller

    def script(self):
        action_js = self._generate_controller_actions(self.controller)
        dependencies = self._get_dependent_imports(self.controller)
        exception_imports, exception_definitions = self._generate_exceptions(
            self.controller
        )

        # Generate imports
        api_import_path = self.get_global_import_path("api.ts")
        controller_import_path = self.get_global_import_path("controllers.ts")
        yield CodeBlock(
            f"import {{ __request, FetchErrorBase }} from '{api_import_path}';",
            f"import type {{ {', '.join(dependencies)} }} from '{controller_import_path}';",
            *exception_imports,
        )

        # Generate actions
        for action in action_js:
            yield CodeBlock(action)

        for exception in exception_definitions:
            yield CodeBlock(exception)

    def _generate_controller_actions(self, controller: ControllerWrapper):
        """
        Generate all actions that are either owned directly by this controller
        or one of the parents, since they'll be separately mounted to this endpoint.

        """
        # Convert each action. We also include the superclass methods, since they're
        # actually bound to the controller instance with separate urls.
        all_actions = [
            ActionInterface.from_action(
                action,
                action.controller_to_url[controller.controller],
                controller.controller,
            )
            for action in controller.all_actions
        ]

        return [typescript_action.to_js() for typescript_action in all_actions]

    def _get_dependent_imports(self, parsed_controller: ControllerWrapper):
        deps = set()
        for action in parsed_controller.all_actions:
            if action.request_body:
                deps.add(action.request_body.name.global_name)
            response_body = action.response_bodies.get(parsed_controller.controller)
            if response_body:
                deps.add(response_body.name.global_name)
        return deps

    def _generate_exceptions(self, parsed_controller: ControllerWrapper):
        """Wrapper around the model to create a concrete Exception class"""
        controllers_import_path = self.get_global_import_path("controllers.ts")
        embedded_types = ControllerWrapper.get_all_embedded_types(
            [parsed_controller], include_superclasses=True
        )

        imports = [
            f"import {{ {exception.name.global_name} as {self._get_exception_import_name(exception)} }} from '{controllers_import_path}';"
            for exception in embedded_types.exceptions
        ]

        definitions = [
            f"export class {exception.name.local_name} extends FetchErrorBase<{self._get_exception_import_name(exception)}> {{}}"
            for exception in embedded_types.exceptions
        ]

        return imports, definitions

    def _get_exception_import_name(self, exception: ExceptionWrapper):
        return f"{exception.name.local_name}Base"


class LocalModelGenerator(LocalGeneratorBase):
    """
    Re-export the globally defined models that are owned (or used) by this
    given controller. This makes IDE type introspection easier for the current
    project file by isolating the models that are involved. We also add alias
    definitions to these imports so even if we were forced to rename the models
    in the global state they'll be the expected definitions here.

    These should mirror the code class names 1:1, with the exception of any
    classes that are both used by the controller and share the same name
    in the global space across different imported files.

    """

    def __init__(
        self,
        controller: ControllerWrapper,
        *,
        managed_path: ManagedViewPath,
        global_root: ManagedViewPath,
    ):
        super().__init__(managed_path=managed_path, global_root=global_root)
        self.controller = controller

    def script(self):
        controller_import_path = self.get_global_import_path("controllers.ts")

        controllers = ControllerWrapper.get_all_embedded_controllers([self.controller])
        embedded_types = ControllerWrapper.get_all_embedded_types(
            [self.controller], include_superclasses=True
        )

        yield CodeBlock(
            *[
                f"export type {{ {controller.name.global_name} as {controller.name.local_name} }} from '{controller_import_path}';"
                for controller in controllers
            ]
        )

        yield CodeBlock(
            *[
                f"export type {{ {model.name.global_name} as {model.name.local_name} }} from '{controller_import_path}';"
                for model in embedded_types.models
            ]
        )

        yield CodeBlock(
            *[
                f"export {{ {enum.name.global_name} as {enum.name.local_name} }} from '{controller_import_path}';"
                for enum in embedded_types.enums
            ]
        )


class LocalUseServerGenerator(LocalGeneratorBase):
    def __init__(
        self,
        controller: ControllerWrapper,
        *,
        managed_path: ManagedViewPath,
        global_root: ManagedViewPath,
    ):
        super().__init__(managed_path=managed_path, global_root=global_root)
        self.controller = controller

    def script(self):
        if not self.controller.render:
            return

        api_import_path = self.get_global_import_path("api.ts")
        links_import_path = self.get_global_import_path("links.ts")
        yield CodeBlock(
            "import React, { useState } from 'react';",
            f"import {{ applySideEffect }} from '{api_import_path}';",
            f"import LinkGenerator from '{links_import_path}';",
        )

        # Verify in this parent function that render is a non-None value
        render_model_name = self.controller.render.name.global_name

        yield from self._generate_imports(self.controller, render_model_name)
        yield from self._generate_interface(self.controller, render_model_name)
        yield from self._generate_hook(self.controller, render_model_name)

    def _generate_imports(self, controller: ControllerWrapper, render_model: str):
        """Generate import statements"""
        controller_import_path = self.get_global_import_path("controllers.ts")
        imports = [
            f"import {{ {render_model}, {controller.name.global_name} }} from '{controller_import_path}';",
        ]

        if controller.all_actions:
            imports.append(
                f"import {{ {', '.join(action.name for action in controller.all_actions)} }} from './actions';"
            )

        yield CodeBlock(*imports)

    def _generate_interface(self, controller: ControllerWrapper, render_model: str):
        """Generate ServerState interface"""
        yield CodeBlock("declare global {", "  var SERVER_DATA: any;", "}")

        yield CodeBlock(
            f"export interface ServerState extends {render_model}, {controller.name.global_name} {{",
            "  linkGenerator: typeof LinkGenerator;",
            "}",
        )

    def _generate_hook(self, controller: ControllerWrapper, render_model: str):
        """Generate useServer hook implementation"""
        server_response = {
            TSLiteral("...serverState"): TSLiteral("...serverState"),
            "linkGenerator": TSLiteral("LinkGenerator"),
        }

        for action in controller.all_actions:
            server_response[TSLiteral(action.name)] = (
                TSLiteral(f"applySideEffect({action.name}, setControllerState)")
                if action.action_type == FunctionActionType.SIDEEFFECT
                else TSLiteral(action.name)
            )

        response_body = python_payload_to_typescript(server_response)
        # Special case: refactor to an explicit controller property
        server_key = controller.controller.__name__

        optional_model_name = f"{render_model}Optional"
        yield CodeBlock(f"export type {optional_model_name} = Partial<{render_model}>;")

        yield CodeBlock(
            "export const useServer = () : ServerState => {",
            f"  const [serverState, setServerState] = useState(SERVER_DATA['{server_key}'] as {render_model});\n",
            f"  const setControllerState = (payload: {optional_model_name}) => {{",
            "    setServerState((state) => ({",
            "      ...state,",
            "      ...payload,",
            "    }));",
            "  };\n",
            f"  return {response_body}",
            "};",
        )


class LocalIndexGenerator(LocalGeneratorBase):
    def __init__(
        self,
        controller: ControllerWrapper,
        *,
        managed_path: ManagedViewPath,
        global_root: ManagedViewPath,
    ):
        super().__init__(managed_path=managed_path, global_root=global_root)
        self.controller = controller

    def script(self):
        exports = []
        for module in ["actions", "links", "models", "useServer"]:
            module_file = self.managed_path.parent / f"{module}.ts"
            if module_file.exists() and module_file.read_text().strip():
                exports.append(f"export * from './{module}';")

        yield CodeBlock(*exports)



import sys

import pytest

from mountaineer.compat import StrEnum


class GoodStrEnum(StrEnum):
    one = "1"
    two = "2"
    three = bytes("3", "ascii").decode("ascii")
    four = bytes("4", "latin1").decode("latin1")


def test_first_failed_str_enum():
    with pytest.raises(TypeError, match="1 is not a string"):

        class FirstFailedStrEnum(StrEnum):
            one = 1  # type: ignore
            two = "2"


def test_second_failed_str_enum():
    with pytest.raises(TypeError, match="2 is not a string"):

        class SecondFailedStrEnum(StrEnum):
            one = "1"
            two = (2,)  # type: ignore
            three = "3"


def test_third_failed_str_enum():
    with pytest.raises(TypeError, match="2 is not a string"):

        class ThirdFailedStrEnum(StrEnum):
            one = "1"
            two = 2  # type: ignore


def test_third_failed_str_enum_encoding_issue():
    with pytest.raises(
        TypeError, match="encoding must be a string, not %r" % (sys.getdefaultencoding,)
    ):

        class FourthFailedStrEnum(StrEnum):
            one = "1"
            two = b"2", sys.getdefaultencoding  # type: ignore


def test_third_failed_str_enum_errors_issue():
    with pytest.raises(TypeError, match="errors must be a string, not 9"):

        class FifthFailedStrEnum(StrEnum):
            one = "1"
            two = b"2", "ascii", 9  # type: ignore


class DumbMixin:
    def __str__(self):  # type: ignore
        # Overrides the necessary __str__ method in the class handler
        # to prevent the StrEnum from returning the value of the enum
        return "don't do this"


class DumbStrEnum(DumbMixin, StrEnum):  # type: ignore
    five = "5"
    six = "6"
    seven = "7"


def test_dumb_str_enum():
    assert DumbStrEnum.seven == "7"
    assert str(DumbStrEnum.seven) == "don't do this"


class EnumMixin:
    def hello(self):
        pass


class HelloEnum(EnumMixin, StrEnum):
    eight = "8"


def test_hello_enum():
    assert HelloEnum.eight == "8"
    assert HelloEnum.eight == str(HelloEnum.eight)


class GoodbyeMixin:
    def goodbye(self):
        pass


class GoodbyeEnum(GoodbyeMixin, EnumMixin, StrEnum):
    nine = "9"


def test_goodbye_enum():
    assert GoodbyeEnum.nine == "9"
    assert GoodbyeEnum.nine == str(GoodbyeEnum.nine)



from abc import ABC, abstractmethod, abstractproperty
from types import UnionType
from typing import (
    Annotated,
    Any,
    Dict,
    List,
    Literal,
    Set,
    Tuple,
    Union,
    get_args,
    get_origin,
)


def is_union_type(type_: type) -> bool:
    """Check if a type is a Union type"""
    return get_origin(type_) is Union or isinstance(type_, UnionType)


def get_union_types(type_: type) -> list[type]:
    """Get the types from a Union type"""
    if not is_union_type(type_):
        raise ValueError(f"Expected Union type, got {type_}")
    return list(get_args(type_))


def is_none_type(field_type: Any):
    return field_type is None or field_type is type(None)


class TypeDefinition(ABC):
    """Base class for all type definitions"""

    @abstractproperty
    def children(self) -> list[Any]:
        """Return a list of all types this definition wraps"""
        raise NotImplementedError

    @abstractmethod
    def update_children(self, children: list[Any]):
        """Return a new instance with updated children"""
        raise NotImplementedError

    def __repr__(self):
        return f"{self.__class__.__name__}({', '.join(repr(child) for child in self.children)})"


class Or(TypeDefinition):
    """Represents a Union type"""

    types: tuple[Any, ...]

    def __init__(self, *types):
        self.types = types

    @property
    def children(self):
        return list(self.types)

    def update_children(self, children):
        self.types = tuple(children)


class ListOf(TypeDefinition):
    """Represents a List type"""

    type: Any

    def __init__(self, type):
        self.type = type

    @property
    def children(self):
        return [self.type]

    def update_children(self, children):
        assert len(children) == 1
        self.type = children[0]


class DictOf(TypeDefinition):
    """Represents a Dict type"""

    key_type: Any
    value_type: Any

    def __init__(self, key, value):
        self.key_type = key
        self.value_type = value

    @property
    def children(self):
        return [self.key_type, self.value_type]

    def update_children(self, children):
        assert len(children) == 2
        self.key_type = children[0]
        self.value_type = children[1]


class TupleOf(TypeDefinition):
    """Represents a Tuple type"""

    types: tuple[Any, ...]

    def __init__(self, *types):
        self.types = types

    @property
    def children(self):
        return list(self.types)

    def update_children(self, children):
        self.types = tuple(children)


class SetOf(TypeDefinition):
    """Represents a Set type"""

    type: Any

    def __init__(self, type_):
        self.type = type_

    @property
    def children(self):
        return [self.type]

    def update_children(self, children):
        self.types = tuple(children)


class LiteralOf(TypeDefinition):
    """Represents a Literal type"""

    values: list[Any]

    def __init__(self, *values):
        self.values = [value if not is_none_type(value) else None for value in values]
        self._validate_primitive_values(self.values)

    @staticmethod
    def _validate_primitive_values(values: List[Any]) -> None:
        """
        Ensures all values are primitive types (str, int, float, bool, None).
        Raises TypeError for non-primitive values.

        """
        for value in values:
            if not isinstance(value, (str, int, float, bool)) and value is not None:
                raise TypeError(
                    f"Literal values must be primitive types (str, int, float, bool, None). "
                    f"Got {type(value)} for value: {value}"
                )

    @property
    def children(self):
        return []

    def update_children(self, children):
        pass


class TypeParser:
    """
    Type parser that converts Python types into TypeDefinition instances
    """

    def parse_type(self, field_type: Any):
        """
        Convert any Python type into a TypeDefinition instance

        Args:
            field_type: Any Python type

        Returns:
            TypeDefinition: Corresponding TypeDefinition instance
        """
        # Handle None type
        if is_none_type(field_type):
            return type(None)

        # Handle unions
        if is_union_type(field_type):
            union_types = get_union_types(field_type)
            return Or(*[self.parse_type(arg) for arg in union_types])

        # Handle built-in collections
        origin_type = get_origin(field_type)
        if origin_type is not None:
            return self._parse_origin_type(field_type, origin_type)

        # Handle basic types
        return self._parse_basic_type(field_type)

    def _parse_origin_type(self, field_type: Any, origin_type: Any) -> TypeDefinition:
        """Parse types with origin (e.g., List[int], Dict[str, int])"""
        args = get_args(field_type)

        # Handle Annotated types by extracting the first argument (the actual type)
        # and ignoring the metadata (the second and subsequent arguments)
        if origin_type is Annotated:
            if not args:
                raise ValueError(f"Annotated type {field_type} has no arguments")
            # Parse only the first argument (the actual type)
            parsed_type = self.parse_type(args[0])
            # If the parsed type is not a TypeDefinition, wrap it in Or with a single type
            if not isinstance(parsed_type, TypeDefinition):
                return Or(parsed_type)
            return parsed_type

        args = tuple(self.parse_type(arg) for arg in args)

        if origin_type in (list, List):
            return ListOf(type=args[0])

        if origin_type in (dict, Dict):
            return DictOf(key=args[0], value=args[1])

        if origin_type in (tuple, Tuple):
            return TupleOf(*args)

        if origin_type in (set, Set):
            return SetOf(args[0])

        if origin_type is Literal:
            return LiteralOf(*args)

        # For unknown origin types, wrap in Or
        raise ValueError(f"Unsupported origin type: {origin_type}")

    def _parse_basic_type(self, field_type: Any) -> Any:
        """Parse basic types without args"""
        if isinstance(field_type, type):
            if issubclass(field_type, (list, List)):
                return ListOf(Any)
            elif issubclass(field_type, (dict, Dict)):
                return DictOf(key=Any, value=Any)
            elif issubclass(field_type, (tuple, Tuple)):  # type: ignore
                return TupleOf(Any)
            elif issubclass(field_type, (set, Set)):
                return SetOf(Any)

        return field_type



from enum import Enum
from pathlib import Path
from typing import Any, List

import pytest
from fastapi import File
from pydantic import BaseModel

from mountaineer.actions.passthrough_dec import passthrough
from mountaineer.actions.sideeffect_dec import sideeffect
from mountaineer.app import AppController
from mountaineer.client_builder.file_generators.base import CodeBlock
from mountaineer.client_builder.file_generators.locals import (
    LocalActionGenerator,
    LocalGeneratorBase,
    LocalIndexGenerator,
    LocalLinkGenerator,
    LocalModelGenerator,
    LocalUseServerGenerator,
)
from mountaineer.client_builder.parser import (
    ControllerParser,
    ControllerWrapper,
)
from mountaineer.controller import ControllerBase
from mountaineer.paths import ManagedViewPath
from mountaineer.render import RenderBase


# Test Models and Enums
class ExampleStatus(Enum):
    ACTIVE = "active"
    INACTIVE = "inactive"


class ExampleBaseModel(BaseModel):
    name: str
    status: ExampleStatus


class ExampleRequestModel(BaseModel):
    query: str
    limit: int = 10


class ExampleResponseModel(BaseModel):
    results: List[str]
    total: int


class ExampleRenderModel(RenderBase):
    title: str
    items: List[ExampleBaseModel]


# Test Controllers
class ExampleBaseController(ControllerBase):
    @passthrough
    def base_action(self) -> ExampleResponseModel:  # type: ignore
        """Base action that returns a response model"""
        pass


class ExampleController(ExampleBaseController):
    url = "/test"
    view_path = "/test.tsx"

    async def render(  # type: ignore
        self,
        path_param: str,
        query_param: int = 0,
        enum_param: ExampleStatus = ExampleStatus.ACTIVE,
    ) -> ExampleRenderModel:  # type: ignore
        """Main render method"""
        pass

    @passthrough
    def get_data(self) -> ExampleBaseModel:  # type: ignore
        """Get basic data"""
        pass

    @sideeffect
    def update_data(self, data: ExampleRequestModel) -> ExampleResponseModel:  # type: ignore
        """Update data with side effects"""
        pass

    @sideeffect
    async def upload_file(self, file: bytes = File(...)) -> ExampleResponseModel:  # type: ignore
        """File upload endpoint"""
        pass


@pytest.fixture
def managed_path(tmp_path: Path) -> ManagedViewPath:
    controller_path = tmp_path / "test_controller"
    controller_path.mkdir()
    return ManagedViewPath(controller_path)


@pytest.fixture
def global_root(tmp_path: Path) -> ManagedViewPath:
    return ManagedViewPath(tmp_path)


@pytest.fixture
def controller_parser() -> ControllerParser:
    return ControllerParser()


@pytest.fixture
def controller_wrapper(controller_parser: ControllerParser) -> ControllerWrapper:
    app_controller = AppController(view_root=Path())
    app_controller.register(ExampleController())

    return controller_parser.parse_controller(ExampleController)


class TestLocalGeneratorBase:
    @pytest.fixture
    def generator(self, managed_path: ManagedViewPath, global_root: ManagedViewPath):
        class ConcreteGeneratorBase(LocalGeneratorBase):
            def script(self):
                yield CodeBlock()

        return ConcreteGeneratorBase(managed_path=managed_path, global_root=global_root)

    def test_get_global_import_path(self, generator: LocalGeneratorBase) -> None:
        result: str = generator.get_global_import_path("test.ts")
        assert isinstance(result, str)
        assert "../" in result
        assert result.endswith("test")


class TestLocalLinkGenerator:
    @pytest.fixture
    def generator(
        self,
        managed_path: ManagedViewPath,
        global_root: ManagedViewPath,
        controller_wrapper: ControllerWrapper,
    ) -> Any:
        return LocalLinkGenerator(
            controller=controller_wrapper,
            managed_path=managed_path,
            global_root=global_root,
        )

    def test_script_generation(self, generator: LocalLinkGenerator) -> None:
        result = list(generator.script())
        assert len(result) > 0
        content = "\n".join(block.content for block in result)
        assert "import" in content
        assert "getLink" in content
        assert "path_param" in content
        assert "query_param" in content
        assert "enum_param" in content

    def test_get_link_implementation_with_parameters(
        self, generator: LocalLinkGenerator
    ) -> None:
        impl = generator._get_link_implementation(generator.controller)
        assert "path_param" in impl
        assert "query_param?" in impl  # Optional parameter
        assert "enum_param?" in impl  # Optional parameter
        assert "/test" in impl

    def test_get_imports(self, generator: LocalLinkGenerator) -> None:
        imports = list(generator._get_imports(generator.controller))
        assert any("../api" in block.content for block in imports)
        assert any("../controllers" in block.content for block in imports)


class TestLocalActionGenerator:
    @pytest.fixture
    def generator(
        self,
        managed_path: ManagedViewPath,
        global_root: ManagedViewPath,
        controller_wrapper: ControllerWrapper,
    ) -> Any:
        return LocalActionGenerator(
            controller=controller_wrapper,
            managed_path=managed_path,
            global_root=global_root,
        )

    def test_generate_controller_actions(self, generator: LocalActionGenerator) -> None:
        actions = list(generator._generate_controller_actions(generator.controller))
        assert len(actions) == 4  # base_action, get_data, update_data, upload_file
        action_names: set[str] = {
            action
            for action in " ".join(actions).split()
            if action in ["base_action", "get_data", "update_data", "upload_file"]
        }
        assert len(action_names) == 4

    def test_get_dependent_imports(self, generator: LocalActionGenerator) -> None:
        deps = generator._get_dependent_imports(generator.controller)

        # Response wrapped models
        assert deps == {
            "BaseActionResponseWrapped",
            "ExampleRequestModel",
            "GetDataResponseWrapped",
            "UpdateDataResponseWrapped",
            "UploadFileForm",
            "UploadFileResponseWrapped",
        }


class TestLocalModelGenerator:
    @pytest.fixture
    def generator(
        self,
        managed_path: ManagedViewPath,
        global_root: ManagedViewPath,
        controller_wrapper: ControllerWrapper,
    ) -> Any:
        return LocalModelGenerator(
            controller=controller_wrapper,
            managed_path=managed_path,
            global_root=global_root,
        )

    def test_script_generation(self, generator: LocalModelGenerator) -> None:
        result: List[Any] = list(generator.script())
        assert len(result) > 0
        content: str = "\n".join(block.content for block in result)

        # Check for model exports
        assert "export type { ExampleRequestModel as ExampleRequestModel }" in content
        assert "export type { ExampleResponseModel as ExampleResponseModel }" in content
        assert "export type { ExampleRenderModel as ExampleRenderModel }" in content

        # Check for enum exports
        assert "export { ExampleStatus as ExampleStatus }" in content


class TestLocalUseServerGenerator:
    @pytest.fixture
    def generator(
        self,
        managed_path: ManagedViewPath,
        global_root: ManagedViewPath,
        controller_wrapper: ControllerWrapper,
    ) -> Any:
        return LocalUseServerGenerator(
            controller=controller_wrapper,
            managed_path=managed_path,
            global_root=global_root,
        )

    def test_script_generation_with_render(
        self, generator: LocalUseServerGenerator
    ) -> None:
        result: List[Any] = list(generator.script())
        content: str = "\n".join(block.content for block in result)
        assert "useServer" in content
        assert "ServerState" in content
        assert "useState" in content
        assert "applySideEffect" in content


class TestLocalIndexGenerator:
    @pytest.fixture
    def generator(
        self,
        managed_path: ManagedViewPath,
        global_root: ManagedViewPath,
        controller_wrapper: ControllerWrapper,
    ) -> Any:
        return LocalIndexGenerator(
            controller=controller_wrapper,
            managed_path=managed_path,
            global_root=global_root,
        )

    def test_script_generation(
        self, generator: LocalIndexGenerator, managed_path: ManagedViewPath
    ) -> None:
        (managed_path.parent / "actions.ts").write_text(
            "export const action = () => {}"
        )
        (managed_path.parent / "models.ts").write_text("export type Model = {}")

        result: List[Any] = list(generator.script())
        assert len(result) > 0
        content: str = "\n".join(block.content for block in result)
        assert "export * from './actions'" in content
        assert "export * from './models'" in content



from collections import Counter
from typing import Sequence, Type

from inflection import camelize
from pydantic import BaseModel

from mountaineer.client_builder.parser import (
    ControllerParser,
    ControllerWrapper,
    CoreWrapper,
    ModelWrapper,
)
from mountaineer.client_builder.typescript import normalize_interface


class AliasManager:
    """
    Since Python classes are implicitly isolated by {module}.{name} we can have
    multiple definitions with the same class name within our runtime. In our
    generated TypeScript this is not available to us because we use one
    consolidated file to implement the ground truth interfaces.

    While we could automatically prefix each definition with the module, this gets
    to be quite long and cumbersome. Instead, we can use this alias manager
    to detect conflicts and resolve them by renaming the conflicting definitions.

    """

    def assign_global_names(self, parser: ControllerParser):
        """
        Assign globally unique names to potentially duplicate models, enums, controllers, etc

        """
        reference_counts: Counter[str] = Counter()

        # Each of these dictionaries are keyed with the actual classes in memory themselves, so
        # any values should be unique representations of different logical classes
        for self_reference in parser.parsed_self_references:
            self_reference.name = normalize_interface(self_reference.name)
            # No need to update the reference counts, since we expect these to just
            # point to an existing model anyway

        parsed_groups: Sequence[Sequence[CoreWrapper]] = [
            list(parser.parsed_models.values()),
            list(parser.parsed_enums.values()),
            list(parser.parsed_exceptions.values()),
            list(parser.parsed_controllers.values()),
        ]

        for parsed_group in parsed_groups:
            for parsed_wrapper in parsed_group:
                parsed_wrapper.name.global_name = normalize_interface(
                    parsed_wrapper.name.global_name
                )
                reference_counts.update([parsed_wrapper.name.global_name])

        # Any reference counts that have more than one reference need to be uniquified
        duplicate_names = {
            name for name, count in reference_counts.items() if count > 1
        }

        converted_models: dict[Type[BaseModel], str] = {}

        # Models must be updated before the self references
        for parsed_group in parsed_groups:
            for parsed_wrapper in parsed_group:
                if parsed_wrapper.name.global_name in duplicate_names:
                    prefix = self._typescript_prefix_from_module(
                        parsed_wrapper.module_name
                    )
                    parsed_wrapper.name.global_name = (
                        f"{prefix}_{parsed_wrapper.name.global_name}"
                    )

                    if isinstance(parsed_wrapper, ModelWrapper):
                        converted_models[parsed_wrapper.model] = (
                            parsed_wrapper.name.global_name
                        )

        # Only once we update the models should we update the self references to the
        # new values - otherwise the lookup map will be empty
        for self_reference in parser.parsed_self_references:
            if self_reference.model in converted_models:
                self_reference.name = converted_models[self_reference.model]

    def assign_local_names(self, parser: ControllerParser):
        """
        Whereas our global logic has to deal with the possibility of multiple classes with
        the same name across the project, typically there's only one definition of a class
        that's imported into a single controller file. This method will assign unique "local names"
        as a shortcut to make it easier for clients to reference these classes.

        """
        for controller in parser.parsed_controllers.values():
            # This should mirror the same logic that the LocalModelGenerator uses to populate
            # the models.ts file that's tied to each controller
            controllers = ControllerWrapper.get_all_embedded_controllers([controller])
            embedded_types = ControllerWrapper.get_all_embedded_types(
                [controller], include_superclasses=True
            )

            reference_counter: Counter[str] = Counter()

            parsed_groups: Sequence[Sequence[CoreWrapper]] = [
                controllers,
                embedded_types.models,
                embedded_types.enums,
                embedded_types.exceptions,
            ]

            for parsed_group in parsed_groups:
                for parsed_wrapper in parsed_group:
                    parsed_wrapper.name.local_name = normalize_interface(
                        parsed_wrapper.name.local_name
                    )
                    reference_counter.update([parsed_wrapper.name.local_name])

            duplicate_names = {
                name for name, count in reference_counter.items() if count > 1
            }

            for parsed_group in parsed_groups:
                for parsed_wrapper in parsed_group:
                    if parsed_wrapper.name.local_name in duplicate_names:
                        prefix = self._typescript_prefix_from_module(
                            parsed_wrapper.module_name
                        )
                        parsed_wrapper.name.local_name = (
                            f"{prefix}_{parsed_wrapper.name.local_name}"
                        )

    def _typescript_prefix_from_module(self, module: str):
        module_parts = module.split(".")
        module_parts = [module.strip("_") for module in module_parts]
        return "".join([camelize(component) for component in module_parts])



from json import dumps as json_dumps
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest.mock import patch

import pytest

from mountaineer.controller import ControllerBase
from mountaineer.paths import (
    ManagedViewPath,
    generate_relative_import,
    is_path_file,
    resolve_package_path,
)


@pytest.mark.parametrize(
    "path,expected_is_file",
    [
        (
            Path("fake-dir/fileA.js"),
            True,
        ),
        (
            Path("fake-dir/folder"),
            False,
        ),
        (
            Path("fake-dir/.hidden-folder"),
            False,
        ),
        (
            Path("fake-dir/.hidden-file.js"),
            True,
        ),
    ],
)
def test_is_path_file_heuristic(path: Path, expected_is_file: Path):
    """
    By providing paths that don't actually exist, we force our function
    to use the heuristic
    """
    assert not path.exists()
    assert is_path_file(path) == expected_is_file


@pytest.mark.parametrize(
    "current_import,desired_import,strip_js_extensions,expected",
    [
        (Path("src/fileA.js"), Path("src/fileB.js"), True, "./fileB"),
        (
            Path("src/fileA.js"),
            Path("src/subdir/fileB.js"),
            True,
            "./subdir/fileB",
        ),
        (Path("src/subdir/fileA.js"), Path("src/fileB.js"), True, "../fileB"),
        (
            Path("src/subdir/fileA.js"),
            Path("src/folderB/fileB.js"),
            True,
            "../folderB/fileB",
        ),
        (
            Path(
                "/Users/root/projects/mountaineer/ci_webapp/ci_webapp/views/app/home/_server"
            ),
            Path(
                "/Users/root/projects/mountaineer/ci_webapp/ci_webapp/views/_server/server.tsx"
            ),
            True,
            "../../../_server/server",
        ),
        (
            Path("src/subdir/fileA.js"),
            Path("src/folderB/fileB.js"),
            False,
            "../folderB/fileB.js",
        ),
    ],
)
def test_generate_relative_import(
    current_import: Path,
    desired_import: Path,
    strip_js_extensions: bool,
    expected: str,
):
    assert (
        generate_relative_import(
            current_import, desired_import, strip_js_extensions=strip_js_extensions
        )
        == expected
    )


def test_managed_view_path_not_inheriting_root_state():
    """
    Only ManagedViewPaths that are explicitly created with `from_view_root` should
    be considered root views. All other paths should be considered subviews.

    """
    root_path = ManagedViewPath.from_view_root("root_view")
    detail_path = root_path / "detail" / "page.tsx"
    not_root_path = ManagedViewPath("not_root_view")

    # Test root view construction - unless created with the class constructor
    # we don't know that the path is in fact a root view
    assert root_path.is_root_link is True
    assert detail_path.is_root_link is False
    assert not_root_path.is_root_link is False


def test_managed_view_path_inherits_root_pointer():
    """
    Ensure all path functions that create new paths will be of type
    ManagedViewPath instead of reverting to a Path.

    """
    root_path = ManagedViewPath.from_view_root("root_view")
    subpath = root_path / "subdir" / "file.tsx"

    def common_assert_path_type(path):
        assert isinstance(path, ManagedViewPath)
        assert path.is_root_link is False
        assert path.root_link == root_path

    # Test all path functions maintain the custom path class
    common_assert_path_type(subpath)
    common_assert_path_type(subpath.resolve())
    common_assert_path_type(subpath.absolute())
    common_assert_path_type(subpath.relative_to(root_path))
    common_assert_path_type(subpath.with_name("new_name.tsx"))
    common_assert_path_type(subpath.with_suffix(".js"))
    common_assert_path_type(subpath.parent)


def test_managed_view_paths_code_directories(tmpdir):
    """
    Ensure that the paths can correctly derive the managed folders.

    """
    root_path = ManagedViewPath.from_view_root(tmpdir)
    assert root_path.get_managed_code_dir() == root_path / "_server"
    assert root_path.get_managed_static_dir() == root_path / "_static"
    assert root_path.get_managed_ssr_dir() == root_path / "_ssr"

    # Non-root paths should only yield the server directory
    non_root_path = root_path / "subdir"
    non_root_path.mkdir()

    assert non_root_path.get_managed_code_dir() == root_path / "subdir" / "_server"
    with pytest.raises(ValueError):
        non_root_path.get_managed_static_dir()
    with pytest.raises(ValueError):
        non_root_path.get_managed_ssr_dir()


def test_managed_view_paths_get_controller(tmpdir):
    """
    Ensure that we can get the full view's filesystem path from a controller.
    """
    root_path = ManagedViewPath.from_view_root(tmpdir)

    class ExampleController(ControllerBase):
        view_path = "/detail/page.tsx"

        def render(self) -> None:
            pass

    controller_path = root_path.get_controller_view_path(ExampleController())
    assert controller_path == root_path / "detail/page.tsx"
    assert controller_path.is_root_link is False
    assert controller_path.root_link == root_path


def test_copy():
    path = ManagedViewPath.from_view_root("root_view", package_root_link="package_root")
    new_path = path.copy()

    assert id(path) != id(new_path)
    assert str(path.root_link) == str(new_path.root_link)
    assert str(path.package_root_link) == str(new_path.package_root_link)


@pytest.mark.parametrize(
    "package_name, filename",
    [
        ("my-awesome-project", "my_awesome_project.pth"),
        ("MyAwesomeProject", "myawesomeproject.pth"),
    ],
)
def test_resolve_symbolic_links(package_name: str, filename: str):
    with (
        TemporaryDirectory() as site_packages_raw,
        TemporaryDirectory() as package_dir_raw,
    ):
        site_packages_path = Path(site_packages_raw)
        package_dir_path = Path(package_dir_raw)

        # Original code directory
        (package_dir_path / package_name).mkdir()

        # pth files are in txt format and have an explicit link
        # to the root path, not the code package itself
        (site_packages_path / filename).write_text(str(package_dir_path.absolute()))

        with patch("importlib.metadata.distribution") as mock_distribution:
            mock_distribution.return_value.name = package_name
            mock_distribution.return_value.files = [
                site_packages_path / filename,
                site_packages_path / "other_file",
            ]
            mock_distribution.return_value.locate_file.side_effect = lambda x: x

            assert resolve_package_path(package_name) == package_dir_path / package_name


@pytest.mark.parametrize(
    "package_name, egg_info_name",
    [
        (
            "my-awesome-project",
            "my_awesome_project-0.1.0.dist-info",
        ),
        (
            "MyAwesomeProject",
            "MyAwesomeProject-0.1.0.dist-info",
        ),
    ],
)
def test_resolve_dist_links(package_name: str, egg_info_name: str):
    with (
        TemporaryDirectory() as site_packages_raw,
        TemporaryDirectory() as package_dir_raw,
    ):
        site_packages_path = Path(site_packages_raw)
        package_dir_path = Path(package_dir_raw)

        # Original code directory
        (package_dir_path / package_name).mkdir()

        # direct_url.json files are located within the application's egginfo
        # directory within a local venv
        egg_info_path = site_packages_path / egg_info_name
        egg_info_path.mkdir(exist_ok=True)

        (egg_info_path / "direct_url.json").write_text(
            json_dumps(
                {
                    "dir_info": {"editable": True},
                    "url": f"file://{package_dir_path.absolute()}",
                }
            )
        )

        with patch("importlib.metadata.distribution") as mock_distribution:
            mock_distribution.return_value.name = package_name
            mock_distribution.return_value.files = [
                egg_info_path / "direct_url.json",
                site_packages_path / "other_file",
            ]
            mock_distribution.return_value.locate_file.side_effect = lambda x: x

            assert resolve_package_path(package_name) == package_dir_path / package_name
